#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Datasource for configuration options"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from collections import OrderedDict
import datetime
import os
import six

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging
import oslo_messaging as messaging

from congress.cfg_validator import parsing
from congress.cfg_validator import utils
from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.dse2 import dse_node as dse


LOG = logging.getLogger(__name__)

FILE = u'file'
VALUE = u'binding'
OPTION = u'option'
OPTION_INFO = u'option_info'
INT_TYPE = u'int_type'
FLOAT_TYPE = u'float_type'
STR_TYPE = u'string_type'
LIST_TYPE = u'list_type'
RANGE_TYPE = u'range_type'
URI_TYPE = u'uri_type'
IPADDR_TYPE = u'ipaddr_type'
SERVICE = u'service'
HOST = u'host'
MODULE = u'module'
TEMPLATE = u'template'
TEMPLATE_NS = u'template_ns'
NAMESPACE = u'namespace'


class ValidatorDriver(datasource_driver.PollingDataSourceDriver):
    """Driver for the Configuration validation datasource"""

    # pylint: disable=too-many-instance-attributes

    DS_NAME = u'config'

    def __init__(self, name=None, args=None):
        super(ValidatorDriver, self).__init__(self.DS_NAME, args)

        # { template_hash -> {name, namespaces} }
        self.known_templates = {}
        # { namespace_hash -> namespace_name }
        self.known_namespaces = {}
        # set(config_hash)
        self.known_configs = set()
        # { template_hash -> (conf_hash, conf)[] }
        self.templates_awaited_by_config = {}

        self.agent_api = ValidatorAgentClient()

        self.rule_added = False

        if hasattr(self, 'add_rpc_endpoint'):
            self.add_rpc_endpoint(ValidatorDriverEndpoints(self))
        self._init_end_start_poll()

    # pylint: disable=no-self-use
    def get_context(self):
        """context for RPC. To define"""
        return {}

    @staticmethod
    def get_datasource_info():
        """Gives back a standardized description of the datasource"""
        result = {}
        result['id'] = 'config'
        result['description'] = (
            'Datasource driver that allows OS configs retrieval.')
        result['config'] = {
            'poll_time': constants.OPTIONAL,
            'lazy_tables': constants.OPTIONAL}
        return result

    @classmethod
    def get_schema(cls):
        sch = {
            # option value
            VALUE: [
                {'name': 'option_id', 'desc': 'The represented option'},
                {'name': 'file_id',
                 'desc': 'The file containing the assignement'},
                {'name': 'val', 'desc': 'Actual value'}],
            OPTION: [
                {'name': 'id', 'desc': 'Id'},
                {'name': 'namespace', 'desc': ''},
                {'name': 'group', 'desc': ''},
                {'name': 'name', 'desc': ''}, ],
            # options metadata, omitted : dest
            OPTION_INFO: [
                {'name': 'option_id', 'desc': 'Option id'},
                {'name': 'type', 'desc': ''},
                {'name': 'default', 'desc': ''},
                {'name': 'deprecated', 'desc': ''},
                {'name': 'deprecated_reason', 'desc': ''},
                {'name': 'mutable', 'desc': ''},
                {'name': 'positional', 'desc': ''},
                {'name': 'required', 'desc': ''},
                {'name': 'sample_default', 'desc': ''},
                {'name': 'secret', 'desc': ''},
                {'name': 'help', 'desc': ''}],
            HOST: [
                {'name': 'id', 'desc': 'Id'},
                {'name': 'name', 'desc': 'Arbitraty host name'}],
            FILE: [
                {'name': 'id', 'desc': 'Id'},
                {'name': 'host_id', 'desc': 'File\'s host'},
                {'name': 'template',
                 'desc': 'Template specifying the content of the file'},
                {'name': 'name', 'desc': ''}],
            MODULE: [
                {'name': 'id', 'desc': 'Id'},
                {'name': 'base_dir', 'desc': ''},
                {'name': 'module', 'desc': ''}],
            SERVICE: [
                {'name': 'service', 'desc': ''},
                {'name': 'host', 'desc': ''},
                {'name': 'version', 'desc': ''}],
            TEMPLATE: [
                {'name': 'id', 'desc': ''},
                {'name': 'name', 'desc': ''}, ],
            TEMPLATE_NS: [
                {'name': 'template', 'desc': 'hash'},
                {'name': 'namespace', 'desc': 'hash'}],
            NAMESPACE: [
                {'name': 'id', 'desc': ''},
                {'name': 'name', 'desc': ''}],
            INT_TYPE: [
                {'name': 'option_id', 'desc': ''},
                {'name': 'min', 'desc': ''},
                {'name': 'max', 'desc': ''},
                {'name': 'choices', 'desc': ''}, ],
            FLOAT_TYPE: [
                {'name': 'option_id', 'desc': ''},
                {'name': 'min', 'desc': ''},
                {'name': 'max', 'desc': ''}, ],
            STR_TYPE: [
                {'name': 'option_id', 'desc': ''},
                {'name': 'regex', 'desc': ''},
                {'name': 'max_length', 'desc': ''},
                {'name': 'quotes', 'desc': ''},
                {'name': 'ignore_case', 'desc': ''},
                {'name': 'choices', 'desc': ''}, ],
            LIST_TYPE: [
                {'name': 'option_id', 'desc': ''},
                {'name': 'item_type', 'desc': ''},
                {'name': 'bounds', 'desc': ''}, ],
            IPADDR_TYPE: [
                {'name': 'option_id', 'desc': ''},
                {'name': 'version', 'desc': ''}, ],
            URI_TYPE: [
                {'name': 'option_id', 'desc': ''},
                {'name': 'max_length', 'desc': ''},
                {'name': 'schemes', 'desc': ''}, ],
            RANGE_TYPE: [
                {'name': 'option_id', 'desc': ''},
                {'name': 'min', 'desc': ''},
                {'name': 'max', 'desc': ''}, ],
        }

        return sch

    def poll(self):
        LOG.info("%s:: polling", self.name)
        # Initialize published state to a sensible empty state.
        # Avoids races with queries.
        if self.number_of_updates == 0:
            for tablename in set(self.get_schema()):
                self.state[tablename] = set()
                self.publish(tablename, self.state[tablename],
                             use_snapshot=False)
        self.agent_api.publish_templates_hashes(self.get_context())
        self.agent_api.publish_configs_hashes(self.get_context())
        self.last_updated_time = datetime.datetime.now()
        self.number_of_updates += 1

    def process_config_hashes(self, hashes, host):
        """Handles a list of config files hashes and their retrieval.

        If the driver can process the parsing and translation of the config,
        it registers the configs to the driver.

        :param hashes: A list of config files hashes
        :param host: Name of the node hosting theses config files
        """
        LOG.debug('Received configs list from %s' % host)

        for cfg_hash in set(hashes) - self.known_configs:
            config = self.agent_api.get_config(self.get_context(),
                                               cfg_hash, host)
            if self.process_config(cfg_hash, config, host):
                self.known_configs.add(cfg_hash)
                LOG.debug('Config %s from %s registered' % (cfg_hash, host))

    @lockutils.synchronized('validator_process_template_hashes')
    def process_template_hashes(self, hashes, host):
        """Handles a list of template hashes and their retrieval.

        Uses lock to avoid multiple sending of the same data.
        :param hashes: A list of templates hashes
        :param host: Name of the node hosting theses config files
        """

        LOG.debug('Process template hashes from %s' % host)

        for t_h in set(hashes) - set(self.known_templates):
            LOG.debug('Treating template hash %s' % t_h)
            template = self.agent_api.get_template(self.get_context(), t_h,
                                                   host)

            ns_hashes = template['namespaces']

            for ns_hash in set(ns_hashes) - set(self.known_namespaces):
                namespace = self.agent_api.get_namespace(
                    self.get_context(), ns_hash, host)
                self.known_namespaces[ns_hash] = namespace

            self.known_templates[t_h] = template
            for (c_h, config) in self.templates_awaited_by_config.pop(t_h, []):
                if self.process_config(c_h, config, host):
                    self.known_configs.add(c_h)
                    LOG.debug('Config %s from %s registered (late)' %
                              (c_h, host))
        return True

    def translate_service(self, host_id, service, version):
        """Translates a service infos to SERVICE table.

        :param host_id: Host ID, should reference HOST.ID
        :param service: A service name
        :param version: A version name, can be None
        """
        if not host_id or not service:
            return

        service_row = tuple(
            map(utils.cfg_value_to_congress, (service, host_id, version)))
        self.state[SERVICE].add(service_row)

    def translate_host(self, host_id, host_name):
        """Translates a host infos to HOST table.

        :param host_id: Host ID
        :param host_name: A host name
        """
        if not host_id:
            return

        host_row = tuple(
            map(utils.cfg_value_to_congress, (host_id, host_name)))
        self.state[HOST].add(host_row)

    def translate_file(self, file_id, host_id, template_id, file_name):
        """Translates a file infos to FILE table.

        :param file_id: File ID
        :param host_id: Host ID, should reference HOST.ID
        :param template_id: Template ID, should reference TEMPLATE.ID
        """
        if not file_id or not host_id:
            return

        file_row = tuple(
            map(utils.cfg_value_to_congress,
                (file_id, host_id, template_id, file_name)))
        self.state[FILE].add(file_row)

    def translate_template_namespace(self, template_id, name, ns_ids):
        """Translates a template infos and its namespaces infos.

        Modifies tables : TEMPLATE, NAMESPACE and TEMPLATE_NS

        :param template_id: Template ID
        :param name: A template name
        :param ns_ids: List of namespace IDs, defining this template, should
            reference NAMESPACE.ID
        """

        if not template_id:
            return

        template_row = tuple(
            map(utils.cfg_value_to_congress, (template_id, name)))
        self.state[TEMPLATE].add(template_row)

        for ns_h, ns_name in six.iteritems(ns_ids):
            if not ns_h:
                continue

            namespace_row = tuple(map(utils.cfg_value_to_congress,
                                      (ns_h, ns_name)))
            self.state[NAMESPACE].add(namespace_row)

            tpl_ns_row = tuple(
                map(utils.cfg_value_to_congress, (template_id, ns_h)))
            self.state[TEMPLATE_NS].add(tpl_ns_row)

    # pylint: disable=protected-access,too-many-branches
    def translate_type(self, opt_id, cfg_type):
        """Translates a type to the appropriate type table.

        :param opt_id: Option ID, should reference OPTION.ID
        :param cfg_type: An oslo ConfigType for the referenced option
        """

        if not opt_id:
            return

        if isinstance(cfg_type, types.String):
            tablename = STR_TYPE
            # oslo.config 5.2 begins to use a different representation of
            # choices (OrderedDict). We first convert back to simple list to
            # have consistent output regardless of oslo.config version
            if isinstance(cfg_type.choices, OrderedDict):
                choices = list(map(lambda item: item[0],
                                   cfg_type.choices.items()))
            else:
                choices = cfg_type.choices
            row = (cfg_type.regex, cfg_type.max_length, cfg_type.quotes,
                   cfg_type.ignore_case, choices)

        elif isinstance(cfg_type, types.Integer):
            tablename = INT_TYPE
            # oslo.config 5.2 begins to use a different representation of
            # choices (OrderedDict). We first convert back to simple list to
            # have consistent output regardless of oslo.config version
            if isinstance(cfg_type.choices, OrderedDict):
                choices = list(map(lambda item: item[0],
                                   cfg_type.choices.items()))
            else:
                choices = cfg_type.choices
            row = (cfg_type.min, cfg_type.max, choices)

        elif isinstance(cfg_type, types.Float):
            tablename = FLOAT_TYPE
            row = (cfg_type.min, cfg_type.max)

        elif isinstance(cfg_type, types.List):
            tablename = LIST_TYPE
            row = (type(cfg_type.item_type).__name__, cfg_type.bounds)

        elif isinstance(cfg_type, types.IPAddress):
            tablename = IPADDR_TYPE
            if cfg_type.version_checker == cfg_type._check_ipv4:
                version = 4
            elif cfg_type.version_checker == cfg_type._check_ipv6:
                version = 6
            else:
                version = None
            row = (version,)

        elif isinstance(cfg_type, types.URI):
            tablename = URI_TYPE
            row = (cfg_type.max_length, cfg_type.schemes)

        elif isinstance(cfg_type, types.Range):
            tablename = RANGE_TYPE
            row = (cfg_type.min, cfg_type.max)

        else:
            return

        row = (opt_id,) + row

        if isinstance(cfg_type, types.List):
            self.translate_type(opt_id, cfg_type.item_type)

        self.state[tablename].add(
            tuple(map(utils.cfg_value_to_congress, row)))

    def translate_value(self, file_id, option_id, value):
        """Translates a value to the VALUE table.

        If value is a list, a table entry is added for every list item.
        If value is a dict, a table entry is added for every key-value.
        :param file_id: File ID, should reference FILE.ID
        :param option_id: Option ID, should reference OPTION.ID
        :param value: A value, can be None
        """
        if not file_id:
            return

        if not option_id:
            return

        if isinstance(value, list):
            for v_item in value:
                value_row = tuple(
                    map(utils.cfg_value_to_congress,
                        (option_id, file_id, v_item)))
                self.state[VALUE].add(value_row)
        elif isinstance(value, dict):
            for v_key, v_item in six.iteritems(value):
                value_row = tuple(
                    map(utils.cfg_value_to_congress,
                        (option_id, file_id, '%s:%s' % (v_key, v_item))))
                self.state[VALUE].add(value_row)
        else:
            value_row = tuple(
                map(utils.cfg_value_to_congress,
                    (option_id, file_id, value)))
            self.state[VALUE].add(value_row)

    def translate_option(self, option, group_name):
        """Translates an option metadata to datasource tables.

        Modifies tables : OPTION, OPTION_INFO
        :param option: An IdentifiedOpt object
        :param group_name: Associated section name
        """

        if option is None:
            return

        if not group_name:
            return

        option_row = tuple(map(utils.cfg_value_to_congress, (
            option.id_, option.ns_id, group_name, option.name)))

        self.state[OPTION].add(option_row)

        option_info_row = tuple(
            map(utils.cfg_value_to_congress, (
                option.id_,
                type(option.type).__name__,
                option.default,
                option.deprecated_for_removal,
                option.deprecated_reason,
                option.mutable,
                option.positional,
                option.required,
                option.sample_default,
                option.secret,
                option.help)))

        self.state[OPTION_INFO].add(option_info_row)

    def translate_conf(self, conf, file_id):
        """Translates a config manager to the datasource state.

        :param conf: A config manager ConfigOpts, containing the parsed values
         and the options metadata to read them
        :param file_id: Id of the file, which contains the parsed values
        """

        cfg_ns = conf._namespace

        def _do_translation(option, group_name='DEFAULT'):
            option = option['opt']
            # skip options that do not have the required attributes
            # avoids processing built-in options included by oslo.config, which
            # don't have all the needed IdentifiedOpt attributes.
            # see: https://github.com/openstack/oslo.config/commit/5ad89d40210bf5922de62e30b096634cac36da6c#diff-768b817a50237989cacd1a8064b4a8af  # noqa
            for attribute in ['id_', 'name', 'type', 'ns_id']:
                if not hasattr(option, attribute):
                    return

            self.translate_option(option, group_name)

            try:
                value = option._get_from_namespace(cfg_ns, group_name)
                if hasattr(cfg, 'LocationInfo'):
                    value = value[0]
            except KeyError:
                # No value parsed for this option
                return

            self.translate_type(option.id_, option.type)

            try:
                value = parsing.parse_value(option.type, value)
            except (ValueError, TypeError):
                LOG.warning('Value for option %s is not valid : %s' % (
                    option.name, value))

            self.translate_value(file_id, option.id_, value)

        for _, option in six.iteritems(conf._opts):
            _do_translation(option)

        for group_name, identified_group in six.iteritems(conf._groups):
            for _, option in six.iteritems(identified_group._opts):
                _do_translation(option, group_name)

    def process_config(self, file_hash, config, host):
        """Manages all translations related to a config file.

        Publish tables to PE.
        :param file_hash: Hash of the configuration file
        :param config: object representing the configuration
        :param host: Remote host name
        :return: True if config was processed
        """
        try:
            LOG.debug("process_config hash=%s" % file_hash)
            template_hash = config['template']
            template = self.known_templates.get(template_hash, None)
            if template is None:
                waiting = (
                    self.templates_awaited_by_config.get(template_hash, []))
                waiting.append((file_hash, config))
                self.templates_awaited_by_config[template_hash] = waiting
                LOG.debug('Template %s not yet registered' % template_hash)
                return False

            host_id = utils.compute_hash(host)

            namespaces = [self.known_namespaces.get(h, None).get('data', None)
                          for h in template['namespaces']]

            conf = parsing.construct_conf_manager(namespaces)
            parsing.add_parsed_conf(conf, config['data'])

            for tablename in set(self.get_schema()) - set(self.state):
                self.state[tablename] = set()
                self.publish(tablename, self.state[tablename],
                             use_snapshot=False)

            self.translate_conf(conf, file_hash)

            self.translate_host(host_id, host)

            self.translate_service(
                host_id, config['service'], config['version'])

            file_name = os.path.basename(config['path'])
            self.translate_file(file_hash, host_id, template_hash, file_name)

            ns_hashes = {h: self.known_namespaces[h]['name']
                         for h in template['namespaces']}
            self.translate_template_namespace(template_hash, template['name'],
                                              ns_hashes)

            for tablename in self.state:
                self.publish(tablename, self.state[tablename],
                             use_snapshot=True)
            return True
        except KeyError:
            LOG.error('Config %s from %s NOT registered'
                      % (file_hash, host))
            return False


class ValidatorAgentClient(object):
    """RPC Proxy to access the agent from the datasource."""

    def __init__(self, topic=utils.AGENT_TOPIC):
        transport = messaging.get_transport(cfg.CONF)
        target = messaging.Target(exchange=dse.DseNode.EXCHANGE,
                                  topic=topic,
                                  version=dse.DseNode.RPC_VERSION)
        self.client = messaging.RPCClient(transport, target)

    def publish_configs_hashes(self, context):
        """Asks for config hashes"""
        cctx = self.client.prepare(fanout=True)
        return cctx.cast(context, 'publish_configs_hashes')

    def publish_templates_hashes(self, context):
        """Asks for template hashes"""
        cctx = self.client.prepare(fanout=True)
        return cctx.cast(context, 'publish_templates_hashes')

    # block calling thread
    def get_namespace(self, context, ns_hash, server):
        """Retrieves an explicit namespace from a server given a hash. """
        cctx = self.client.prepare(server=server)
        return cctx.call(context, 'get_namespace', ns_hash=ns_hash)

    # block calling thread
    def get_template(self, context, tpl_hash, server):
        """Retrieves an explicit template from a server given a hash"""
        cctx = self.client.prepare(server=server)
        return cctx.call(context, 'get_template', tpl_hash=tpl_hash)

    # block calling thread
    def get_config(self, context, cfg_hash, server):
        """Retrieves a config from a server given a hash"""
        cctx = self.client.prepare(server=server)
        return cctx.call(context, 'get_config', cfg_hash=cfg_hash)


class ValidatorDriverEndpoints(object):
    """RPC endpoint on the datasource driver for use by the agents"""

    def __init__(self, driver):
        self.driver = driver

    # pylint: disable=unused-argument
    def process_templates_hashes(self, context, **kwargs):
        """Process the template hashes received from a server"""
        LOG.debug(
            'Received template hashes from host %s' % kwargs.get('host', ''))

        self.driver.process_template_hashes(**kwargs)

    # pylint: disable=unused-argument
    def process_configs_hashes(self, context, **kwargs):
        """Process the config hashes received from a server"""
        LOG.debug(
            'Received config hashes from host %s' % kwargs.get('host', ''))

        self.driver.process_config_hashes(**kwargs)



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

""" Generation of JSON from oslo config options (marshalling) """
import collections
import json
import logging

from oslo_config import cfg
from oslo_config import generator
from oslo_config import types

LOG = logging.getLogger(__name__)


class OptionJsonEncoder(json.JSONEncoder):
    """Json encoder used to give a unique representation to namespaces"""

    # pylint: disable=protected-access,method-hidden,too-many-branches
    def default(self, o):
        if isinstance(o, cfg.Opt):
            return {
                'kind': type(o).__name__,
                'deprecated_for_removal': o.deprecated_for_removal,
                'short': o.short,
                'name': o.name,
                'dest': o.dest,
                'deprecated_since': o.deprecated_since,
                'required': o.required,
                'sample_default': o.sample_default,
                'deprecated_opts': o.deprecated_opts,
                'positional': o.positional,
                'default': o.default,
                'secret': o.secret,
                'deprecated_reason': o.deprecated_reason,
                'mutable': o.mutable,
                'type': o.type,
                'metavar': o.metavar,
                'advanced': o.advanced,
                'help': o.help
            }
        elif isinstance(o, (types.ConfigType, types.HostAddress)):
            res = {
                'type': type(o).__name__,
            }

            if isinstance(o, types.Number):
                res['max'] = o.max
                res['min'] = o.min
                # When we build back the type in parsing, we can directly use
                # the list of tuples from choices and it will be in a
                # canonical order (not sorted but the order elements were
                # added)
                if isinstance(o.choices, collections.OrderedDict):
                    res['choices'] = list(o.choices.keys())
                else:
                    res['choices'] = o.choices
            if isinstance(o, types.Range):
                res['max'] = o.max
                res['min'] = o.min
            if isinstance(o, types.String):
                if o.regex and hasattr(o.regex, 'pattern'):
                    res['regex'] = o.regex.pattern
                else:
                    res['regex'] = o.regex
                res['max_length'] = o.max_length
                res['quotes'] = o.quotes
                res['ignore_case'] = o.ignore_case
                if isinstance(o.choices, collections.OrderedDict):
                    res['choices'] = list(o.choices.keys())
                else:
                    res['choices'] = o.choices
            if isinstance(o, types.List):
                res['item_type'] = o.item_type
                res['bounds'] = o.bounds
            if isinstance(o, types.Dict):
                res['value_type'] = o.value_type
                res['bounds'] = o.bounds
            if isinstance(o, types.URI):
                res['schemes'] = o.schemes
                res['max_length'] = o.max_length
            if isinstance(o, types.IPAddress):
                if o.version_checker == o._check_ipv4:
                    res['version'] = 4
                elif o.version_checker == o._check_ipv6:
                    res['version'] = 6

            # Remove unused fields
            remove = [k for k, v in res.items() if not v]
            for k in remove:
                del res[k]

            return res
        elif isinstance(o, cfg.DeprecatedOpt):
            return {
                'name': o.name,
                'group': o.group
            }
        elif isinstance(o, cfg.OptGroup):
            return {
                'title': o.title,
                'help': o.help
            }

        # TODO(vmatt): some options (auth_type, auth_section) from
        # keystoneauth1, loaded by keystonemiddleware.auth,
        # are not defined conventionally (stable/ocata).
        elif isinstance(o, type):
            return {
                'type': 'String'
            }

        else:
            return type(o).__name__


# pylint: disable=protected-access
def generate_ns_data(namespace):
    """Generate a json string containing the namespace"""
    groups = generator._get_groups(generator._list_opts([namespace]))
    return OptionJsonEncoder(sort_keys=True).encode(groups)



# Copyright (c) 2017 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import copy
import json
import jsonschema
import os

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
import yaml

from congress.datalog import compile
from congress.db import db_library_policies
from congress.dse2 import data_service
from congress import exception

LOG = logging.getLogger(__name__)


def validate_policy_item(item):
    schema_json = '''
    {
      "id": "PolicyProperties",
      "title": "Policy Properties",
      "type": "object",
      "required": ["name", "rules"],
      "properties": {
        "name": {
          "title": "Policy unique name",
          "type": "string",
          "minLength": 1,
          "maxLength": 255
        },
        "description": {
          "title": "Policy description",
          "type": "string"
        },
        "kind": {
          "title": "Policy kind",
          "type": "string",
          "enum": ["database", "nonrecursive", "action", "z3", "materialized",
                   "delta", "datasource"]
        },
        "abbreviation": {
          "title": "Policy name abbreviation",
          "type": "string",
          "minLength": 1,
          "maxLength": 5
        },
        "rules": {
          "title": "collection of rules",
          "type": "array",
          "items": {
            "title": "Policy rule",
            "type": "object",
            "required": ["rule"],
            "properties": {
              "rule": {
                "title": "Rule definition following policy grammar",
                "type": "string"
              },
              "name": {
                "title": "User-friendly name",
                "type": "string",
                "maxLength": 255
              },
              "comment": {
                "title": "User-friendly comment",
                "type": "string",
                "maxLength": 255
              }
            }
          }
        }
      }
    }
    '''
    try:
        jsonschema.validate(item, json.loads(schema_json))
    except jsonschema.exceptions.ValidationError as ve:
        raise exception.InvalidPolicyInput(data=str(ve))


class LibraryService (data_service.DataService):
    def __init__(self, name):
        data_service.DataService.__init__(self, name)
        self.name = name
        self.add_rpc_endpoint(DseLibraryServiceEndpoints(self))

    def create_policy(self, policy_dict):
        policy_dict = copy.deepcopy(policy_dict)
        validate_policy_item(policy_dict)
        policy_name = policy_dict['name']

        # check name is valid
        if not compile.string_is_servicename(policy_name):
            raise exception.PolicyException(
                'name `%s` is not a valid policy name' % policy_name)

        # make defaults
        if 'kind' not in policy_dict:
            policy_dict['kind'] = 'nonrecursive'
        if 'abbreviation' not in policy_dict:
            policy_dict['abbreviation'] = policy_name[:5]
        if 'description' not in policy_dict:
            policy_dict['description'] = ''

        try:
            # Note(thread-safety): blocking call
            policy = db_library_policies.add_policy(policy_dict=policy_dict)
            return policy.to_dict()
        except db_exc.DBError:
            LOG.exception('Creating a new library policy failed due to '
                          'backend database error.')
            raise

    def get_policies(self, include_rules=True):
        return [p.to_dict(include_rules)
                for p in db_library_policies.get_policies()]

    def get_policy(self, id_, include_rules=True):
        # Note(thread-safety): blocking call
        policy = db_library_policies.get_policy(id_)
        return policy.to_dict(include_rules)

    def get_policy_by_name(self, name, include_rules=True):
        # Note(thread-safety): blocking call
        policy = db_library_policies.get_policy_by_name(name)
        return policy.to_dict(include_rules)

    def delete_all_policies(self):
        # Note(thread-safety): blocking call
        db_library_policies.delete_policies()

    def delete_policy(self, id_):
        # Note(thread-safety): blocking call
        db_object = db_library_policies.get_policy(id_)
        db_library_policies.delete_policy(id_)
        return db_object.to_dict(include_rules=True)

    def replace_policy(self, id_, policy_dict):
        validate_policy_item(policy_dict)
        policy_name = policy_dict['name']

        # check name is valid
        if not compile.string_is_servicename(policy_name):
            raise exception.PolicyException(
                "Policy name %s is not a valid service name" % policy_name)

        # make defaults
        if 'kind' not in policy_dict:
            policy_dict['kind'] = 'nonrecursive'
        if 'abbreviation' not in policy_dict:
            policy_dict['abbreviation'] = policy_name[:5]
        if 'description' not in policy_dict:
            policy_dict['description'] = ''

        # Note(thread-safety): blocking call
        policy = db_library_policies.replace_policy(
            id_, policy_dict=policy_dict)
        return policy.to_dict()

    def load_policies_from_files(self):
        '''load library policies from library directory

        return total number of files on which error encountered
        '''
        def _load_library_policy_file(full_path):
            try:
                success_policy_count = 0
                error_policy_count = 0
                doc_num_in_file = 0
                file_error = False
                with open(full_path, "r") as stream:
                    policies = yaml.load_all(stream)
                    for policy in policies:
                        try:
                            doc_num_in_file += 1
                            self.create_policy(policy)
                            success_policy_count += 1
                        except db_exc.DBDuplicateEntry:
                            LOG.debug(
                                'Library policy %s (number %s in file %s) '
                                'already exists (likely loaded by another '
                                'Congress instance). Skipping.',
                                policy.get('name', '[no name]'),
                                doc_num_in_file, full_path)
                        except exception.CongressException:
                            LOG.exception(
                                'Library policy %s could not be loaded. '
                                'Skipped. YAML reproduced here %s',
                                policy.get('name', '[no name]'),
                                yaml.dump(policy))
                            error_policy_count += 1
            except Exception:
                LOG.exception(
                    'Failed to load library policy file %s', full_path)
                file_error = True
            return success_policy_count, file_error or (error_policy_count > 0)
        file_count = 0
        file_error_count = 0
        policy_count = 0
        for (dirpath, dirnames, filenames) in os.walk(
                cfg.CONF.policy_library_path):
            for filename in filenames:
                name, extension = os.path.splitext(filename)
                if extension in ['.yaml', '.yml']:
                    count, has_error = _load_library_policy_file(
                        os.path.join(dirpath, filename))
                    if count > 0:
                        file_count += 1
                        policy_count += count
                    if has_error:
                        file_error_count += 1
        LOG.debug(
            '%s library policies from %s files successfully loaded; '
            'error encountered on %s files.',
            policy_count, file_count, file_error_count)
        return file_error_count


class DseLibraryServiceEndpoints(object):
    """RPC endpoints exposed by LibraryService."""

    def __init__(self, data_service):
        self.data_service = data_service

    def create_policy(self, context, policy_dict):
        return self.data_service.create_policy(policy_dict)

    def get_policies(self, context, include_rules=True):
        return self.data_service.get_policies(include_rules)

    def get_policy(self, context, id_, include_rules=True):
        return self.data_service.get_policy(id_, include_rules)

    def get_policy_by_name(self, context, name, include_rules=True):
        return self.data_service.get_policy_by_name(name, include_rules)

    def delete_all_policies(self, context):
        return self.data_service.delete_all_policies()

    def delete_policy(self, context, id_):
        return self.data_service.delete_policy(id_)

    def replace_policy(self, context, id_, policy_dict):
        return self.data_service.replace_policy(id_, policy_dict)



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Unmarshaling of options sent by the agent."""
import inspect
import sys

from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging
from oslo_serialization import jsonutils as json
import six

from congress.cfg_validator import utils

LOG = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class IdentifiedOpt(cfg.Opt):
    """A subclass of option that adds a unique id and a namespace id

    ids are based on hashes
    """
    def __init__(self, id_, ns_id, **kwargs):
        super(IdentifiedOpt, self).__init__(**kwargs)
        self.id_ = id_
        self.ns_id = ns_id


def parse_value(cfgtype, value):
    """Parse and validate a value's type, raising error if check fails.

    :raises: ValueError, TypeError
    """
    return cfgtype(value)


def make_type(type_descr):
    """Declares a new type

    :param type_descr: a type description read from json.
    :return: an oslo config type
    """
    type_name = type_descr['type']
    type_descr = dict(type_descr)
    del type_descr['type']

    if 'item_type' in type_descr:
        item_type = make_type(type_descr['item_type'])
        type_descr['item_type'] = item_type

    if 'value_type' in type_descr:
        value_type = make_type(type_descr['value_type'])
        type_descr['value_type'] = value_type

    return getattr(types, type_name)(**type_descr)


# This function must never fail even if the content/metadata
# of the option were weird.
# pylint: disable=broad-except
def make_opt(option, opt_hash, ns_hash):
    """Declares a new group

    :param name: an option retrieved from json.
    :param opt_hash: the option hash
    :param ns_hash: the hash of the namespace defining it.
    :return: an oslo config option representation augmented with the hashes.
    """
    name = option.get('name', None)
    deprecateds = []

    if option.get('deprecated_opts', None):
        for depr_descr in option.get('deprecated_opts', {}):
            depr_name = depr_descr.get('name', None)
            if depr_name is None:
                depr_name = name
            depr_opt = cfg.DeprecatedOpt(depr_name,
                                         depr_descr.get('group', None))
            deprecateds.append(depr_opt)

    cfgtype = make_type(option.get('type', {}))
    default = option.get('default', None)
    if default:
        try:
            default = cfgtype(default)
        except Exception:
            _, err, _ = sys.exc_info()
            LOG.error('Unvalid default value (%s, %s): %s'
                      % (name, default, err))
    try:
        cfgopt = IdentifiedOpt(
            id_=opt_hash,
            ns_id=ns_hash,
            name=name,
            type=cfgtype,
            dest=option.get('dest', None),
            default=default,
            positional=option.get('positional', None),
            help=option.get('help', None),
            secret=option.get('secret', None),
            required=option.get('required', None),
            sample_default=option.get('sample_default', None),
            deprecated_for_removal=option.get('deprecated_for_removal', None),
            deprecated_reason=option.get('deprecated_reason', None),
            deprecated_opts=deprecateds,
            mutable=option.get('mutable', None))
    except Exception:
        cfgopt = None
        _, err, _ = sys.exc_info()
        LOG.error('Unvalid option definition (%s in %s): %s'
                  % (name, ns_hash, err))
    return cfgopt


def make_group(name, title, help_msg):
    """Declares a new group

    :param name: group name
    :param title: group title
    :param help_msg: descriptive help message
    :return: an oslo config group representation or None for default.
    """
    if name == 'DEFAULT':
        return None

    return cfg.OptGroup(name=name, title=title, help=help_msg)


def add_namespace(conf, ns_dict, ns_hash):
    """Add options from a kind to an already existing config"""

    for group_name, group in six.iteritems(ns_dict):

        try:
            title = group['object'].get('title', None)
            help_msg = group['object'].get('help', None)
        except AttributeError:
            title = help_msg = None
        cfggroup = make_group(group_name, title, help_msg)

        # Get back the instance already stored or register the group.
        if cfggroup is not None:
            # pylint: disable=protected-access
            cfggroup = conf._get_group(cfggroup, autocreate=True)

        for namespace in group['namespaces']:

            for option in namespace[1]:
                opt_hash = utils.compute_hash(ns_hash, group_name,
                                              option['name'])
                cfgopt = make_opt(option, opt_hash, ns_hash)
                conf.register_opt(cfgopt, cfggroup)


def construct_conf_manager(namespaces):
    """Construct a config manager from a list of namespaces data.

    Register options of given namespaces into a cfg.ConfigOpts object.
    A namespaces dict is typically cfg_validator.generator output. Options are
    provided an hash as an extra field.

    :param namespaces: A list of dict, containing options metadata.
    :return: A cfg.ConfigOpts.
    """
    conf = cfg.ConfigOpts()

    for ns_dict in namespaces:
        ns_hash = utils.compute_hash(json.dumps(ns_dict, sort_keys=True))
        add_namespace(conf, ns_dict, ns_hash)

    return conf


def add_parsed_conf(conf, normalized):
    """Add a normalized values container to a config manager.

    :param conf: A cfg.ConfigOpts object.
    :param normalized: A normalized values container, as introduced by oslo
        cfg._Namespace.
    """
    if conf:
        # pylint: disable=protected-access
        conf._namespace = cfg._Namespace(conf)

        # oslo.config version 6.0.1 added extra arg to _add_parsed_config_file
        # we determine the number of args required to use appropriately
        if six.PY2:
            _add_parsed_config_file_args_len = len(inspect.getargspec(
                conf._namespace._add_parsed_config_file).args) - 1
            # - 1 to not count the first param self
        else:
            _add_parsed_config_file_args_len = len(inspect.signature(
                conf._namespace._add_parsed_config_file).parameters)
        if _add_parsed_config_file_args_len == 3:  # oslo.config>=6.0.1
            conf._namespace._add_parsed_config_file(
                '<memory>', [], normalized[0])
        else:
            conf._namespace._add_parsed_config_file([], normalized[0])


def parse_config_file(namespaces, path):
    """Parse a config file from its pre-loaded namespaces.

    :param namespaces: A list of dict, containing namespaces data.
    :param path: Path to the configuration file to parse.
    :return:
    """
    conf = construct_conf_manager(namespaces)
    # pylint: disable=protected-access
    conf._namespace = cfg._Namespace(conf)
    cfg.ConfigParser._parse_file(path, conf._namespace)

    return conf



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import json
import six

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_messaging import exceptions as messaging_exceptions
from oslo_messaging.rpc import dispatcher
from oslo_utils import strutils
from oslo_utils import uuidutils
import pkg_resources
import stevedore

from congress.datalog import compile as datalog_compile
from congress.datasources import constants
from congress.db import datasources as datasources_db
from congress.dse2 import control_bus
from congress import exception


LOG = logging.getLogger(__name__)


class DseNode(object):
    """Addressable entity participating on the DSE message bus.

    The Data Services Engine (DSE) is comprised of one or more DseNode
    instances that each may run one or more DataService instances.  All
    communication between data services uses the DseNode interface.

    Attributes:

    - node_id: The unique ID of this node on the DSE.
    - messaging_config: Configuration options for the message bus.  See
      oslo.messaging for more details.
    - node_rpc_endpoints: List of object instances exposing a remotely
      invokable interface.
    """
    RPC_VERSION = '1.0'
    EXCHANGE = 'congress'
    CONTROL_TOPIC = 'congress-control'
    SERVICE_TOPIC_PREFIX = 'congress-service-'

    loaded_drivers = {}

    def node_rpc_target(self, namespace=None, server=None, fanout=False):
        return messaging.Target(exchange=self.EXCHANGE,
                                topic=self._add_partition(self.CONTROL_TOPIC),
                                version=self.RPC_VERSION,
                                namespace=namespace,
                                server=server,
                                fanout=fanout)

    def service_rpc_target(self, service_id, namespace=None, server=None,
                           fanout=False):
        topic = self._add_partition(self.SERVICE_TOPIC_PREFIX + service_id)
        return messaging.Target(exchange=self.EXCHANGE,
                                topic=topic,
                                version=self.RPC_VERSION,
                                namespace=namespace,
                                server=server,
                                fanout=fanout)

    def _add_partition(self, topic, partition_id=None):
        """Create a seed-specific version of an oslo-messaging topic."""
        partition_id = partition_id or self.partition_id
        if partition_id is None:
            return topic
        return topic + "-" + str(partition_id)

    def __init__(self, messaging_config, node_id, node_rpc_endpoints,
                 partition_id=None):
        self.messaging_config = messaging_config
        self.node_id = node_id
        self.node_rpc_endpoints = node_rpc_endpoints
        # unique identifier shared by all nodes that can communicate
        self.partition_id = partition_id or cfg.CONF.dse.bus_id or "bus"
        self.node_rpc_endpoints.append(DseNodeEndpoints(self))
        self._running = False
        self._services = []
        # uuid to help recognize node_id clash
        self.instance = uuidutils.generate_uuid()
        # TODO(dse2): add detection and logging/rectifying for node_id clash?
        access_policy = dispatcher.DefaultRPCAccessPolicy
        self.context = self._message_context()
        self.transport = messaging.get_rpc_transport(
            self.messaging_config,
            allowed_remote_exmods=[exception.__name__, dispatcher.__name__,
                                   db_exc.__name__, ])
        self._rpctarget = self.node_rpc_target(self.node_id, self.node_id)
        self._rpc_server = messaging.get_rpc_server(
            self.transport, self._rpctarget, self.node_rpc_endpoints,
            executor='eventlet', access_policy=access_policy)

        # # keep track of what publisher/tables local services subscribe to
        # subscribers indexed by publisher and table:
        # {publisher_id ->
        #     {table_name -> set_of_subscriber_ids}}
        self.subscriptions = {}

        # Note(ekcs): A little strange that _control_bus starts before self?
        self._control_bus = control_bus.DseNodeControlBus(self)
        self.register_service(self._control_bus)
        self.periodic_tasks = None
        self.sync_thread = None
        self.start()

    def __del__(self):
        self.stop()
        self.wait()

    def __repr__(self):
        return self.__class__.__name__ + "<%s>" % self.node_id

    def _message_context(self):
        return {'node_id': self.node_id, 'instance': str(self.instance)}

    # Note(thread-safety): blocking function
    def register_service(self, service):
        assert service.node is None
        if self.service_object(service.service_id):
            msg = ('Service %s already exsists on the node %s'
                   % (service.service_id, self.node_id))
            raise exception.DataServiceError(msg)
        access_policy = dispatcher.DefaultRPCAccessPolicy
        service.node = self
        self._services.append(service)
        service._target = self.service_rpc_target(service.service_id,
                                                  server=self.node_id)
        service._rpc_server = messaging.get_rpc_server(
            self.transport, service._target, service.rpc_endpoints(),
            executor='eventlet', access_policy=access_policy)

        if self._running:
            service.start()

        LOG.debug('<%s> Service %s RPC Server listening on %s',
                  self.node_id, service.service_id, service._target)

    # Note(thread-safety): blocking function
    def unregister_service(self, service_id=None, uuid_=None):
        """Unregister service from DseNode matching on service_id or uuid\_

        Only one should be supplied. No-op if no matching service found.
        """
        LOG.debug("unregistering service %s on node %s", service_id,
                  self.node_id)
        service = self.service_object(service_id=service_id, uuid_=uuid_)
        if service is not None:
            self._services.remove(service)
            service.stop()
            # Note(thread-safety): blocking call
            service.wait()

    def get_services(self, hidden=False):
        """Return all local service objects."""
        if hidden:
            return self._services
        return [s for s in self._services if s.service_id[0] != '_']

    def get_global_service_names(self, hidden=False):
        """Return names of all services on all nodes."""
        services = self.get_services(hidden=hidden)
        local_services = [s.service_id for s in services]
        # Also, check services registered on other nodes
        peer_nodes = self.dse_status()['peers']
        peer_services = []
        for node in peer_nodes.values():
            peer_services.extend(
                [srv['service_id'] for srv in node['services']])
        return set(local_services + peer_services)

    def service_object(self, service_id=None, uuid_=None):
        """Return the service object requested.

        Search by service_id or uuid\_ (only one should be supplied).
        None if not found.
        """
        if service_id is not None:
            if uuid_ is not None:
                raise TypeError('service_object() cannot accept both args '
                                'service_id and uuid_')
            for s in self._services:
                if s.service_id == service_id:
                    return s
        elif uuid_ is not None:
            for s in self._services:
                if getattr(s, 'ds_id', None) == uuid_:
                    return s
        else:
            raise TypeError('service_object() requires service_id or '
                            'uuid_ argument, but neither is given.')
        return None

    def start(self):
        LOG.debug("<%s> DSE Node '%s' starting with %s sevices...",
                  self.node_id, self.node_id, len(self._services))

        # Start Node RPC server
        self._rpc_server.start()
        LOG.debug('<%s> Node RPC Server listening on %s',
                  self.node_id, self._rpctarget)

        # Start Service RPC server(s)
        for s in self._services:
            s.start()
            LOG.debug('<%s> Service %s RPC Server listening on %s',
                      self.node_id, s.service_id, s._target)

        self._running = True

    def stop(self):
        if self._running is False:
            return

        LOG.info("Stopping DSE node '%s'", self.node_id)
        for s in self._services:
            s.stop()
        self._rpc_server.stop()
        self._running = False

    # Note(thread-safety): blocking function
    def wait(self):
        for s in self._services:
            # Note(thread-safety): blocking call
            s.wait()
        # Note(thread-safety): blocking call
        self._rpc_server.wait()

    def dse_status(self):
        """Return latest observation of DSE status."""
        return self._control_bus.dse_status()

    def is_valid_service(self, service_id):
        return service_id in self.get_global_service_names(hidden=True)

    # Note(thread-safety): blocking function
    def invoke_node_rpc(self, node_id, method, kwargs=None, timeout=None):
        """Invoke RPC method on a DSE Node.

        :param: node_id: The ID of the node on which to invoke the call.
        :param: method: The method name to call.
        :param: kwargs: A dict of method arguments.

        :returns: The result of the method invocation.

        :raises: MessagingTimeout, RemoteError, MessageDeliveryFailure
        """
        if kwargs is None:
            kwargs = {}
        target = self.node_rpc_target(server=node_id)
        LOG.trace("<%s> Invoking RPC '%s' on %s", self.node_id, method, target)
        client = messaging.RPCClient(self.transport, target, timeout=timeout)
        return client.call(self.context, method, **kwargs)

    # Note(thread-safety): blocking function
    def broadcast_node_rpc(self, method, kwargs=None):
        """Invoke RPC method on all DSE Nodes.

        :param: method: The method name to call.
        :param: kwargs: A dict of method arguments.

        :returns: None
                  Methods are invoked asynchronously and results are dropped.

        :raises: RemoteError, MessageDeliveryFailure
        """
        if kwargs is None:
            kwargs = {}
        target = self.node_rpc_target(fanout=True)
        LOG.trace("<%s> Casting RPC '%s' on %s", self.node_id, method, target)
        client = messaging.RPCClient(self.transport, target)
        client.cast(self.context, method, **kwargs)

    # Note(thread-safety): blocking function
    def invoke_service_rpc(
            self, service_id, method, kwargs=None, timeout=None, local=False,
            retry=None):
        """Invoke RPC method on a DSE Service.

        :param: service_id: The ID of the data service on which to invoke the
            call.
        :param: method: The method name to call.
        :param: kwargs: A dict of method arguments.

        :returns: The result of the method invocation.

        :raises: MessagingTimeout, RemoteError, MessageDeliveryFailure,
                 NotFound
        """
        target = self.service_rpc_target(
            service_id, server=(self.node_id if local else None))
        LOG.trace("<%s> Preparing to invoking RPC '%s' on %s",
                  self.node_id, method, target)
        client = messaging.RPCClient(self.transport, target, timeout=timeout,
                                     retry=retry)
        if not self.is_valid_service(service_id):
            try:
                # First ping the destination to fail fast if unresponsive
                LOG.trace("<%s> Checking responsiveness before invoking RPC "
                          "'%s' on %s", self.node_id, method, target)
                client.prepare(timeout=cfg.CONF.dse.ping_timeout).call(
                    self.context, 'ping')
            except (messaging_exceptions.MessagingTimeout,
                    messaging_exceptions.MessageDeliveryFailure):
                msg = "service '%s' could not be found"
                raise exception.NotFound(msg % service_id)
        if kwargs is None:
            kwargs = {}
        try:
            LOG.trace(
                "<%s> Invoking RPC '%s' on %s", self.node_id, method, target)
            result = client.call(self.context, method, **kwargs)
        except dispatcher.NoSuchMethod:
            msg = "Method %s not supported for datasource %s"
            LOG.exception(msg, method, service_id)
            raise exception.BadRequest(msg % (method, service_id))
        except (messaging_exceptions.MessagingTimeout,
                messaging_exceptions.MessageDeliveryFailure):
            msg = "Request to service '%s' timed out"
            raise exception.Unavailable(msg % service_id)
        LOG.trace("<%s> RPC call returned: %s", self.node_id, result)
        return result

    # Note(thread-safety): blocking function
    def broadcast_service_rpc(self, service_id, method, kwargs=None):
        """Invoke RPC method on all instances of service_id.

        :param: service_id: The ID of the data service on which to invoke the
                call.
        :param: method: The method name to call.
        :param: kwargs: A dict of method arguments.

        :returns: None - Methods are invoked asynchronously and results are
                  dropped.

        :raises: RemoteError, MessageDeliveryFailure
        """
        if kwargs is None:
            kwargs = {}
        if not self.is_valid_service(service_id):
            msg = "service '%s' is not a registered service"
            raise exception.NotFound(msg % service_id)

        target = self.service_rpc_target(service_id, fanout=True)
        LOG.trace("<%s> Casting RPC '%s' on %s", self.node_id, method, target)
        client = messaging.RPCClient(self.transport, target)
        client.cast(self.context, method, **kwargs)

    # Note(ekcs): non-sequenced publish retained to simplify rollout of dse2
    #   to be replaced by handle_publish_sequenced
    # Note(thread-safety): blocking function
    def publish_table(self, publisher, table, data):
        """Invoke RPC method on all insances of service_id.

        :param: service_id: The ID of the data service on which to invoke the
            call.
        :param: method: The method name to call.
        :param: kwargs: A dict of method arguments.

        :returns: None
                  Methods are invoked asynchronously and results are dropped.

        :raises: RemoteError, MessageDeliveryFailure
        """
        LOG.trace("<%s> Publishing from '%s' table %s: %s",
                  self.node_id, publisher, table, data)
        self.broadcast_node_rpc(
            "handle_publish",
            {'publisher': publisher, 'table': table, 'data': data})

    # Note(thread-safety): blocking function
    def publish_table_sequenced(
            self, publisher, table, data, is_snapshot, seqnum):
        """Invoke RPC method on all insances of service_id.

        :param: service_id: The ID of the data service on which to invoke the
            call.
        :param: method: The method name to call.
        :param: kwargs: A dict of method arguments.

        :returns: None
                  Methods are invoked asynchronously and results are dropped.

        :raises: RemoteError, MessageDeliveryFailure
        """
        LOG.trace("<%s> Publishing from '%s' table %s: %s",
                  self.node_id, publisher, table, data)
        self.broadcast_node_rpc(
            "handle_publish_sequenced",
            {'publisher': publisher, 'table': table,
             'data': data, 'is_snapshot': is_snapshot, 'seqnum': seqnum})

    def table_subscribers(self, publisher, table):
        """List services on this node that subscribes to publisher/table."""
        return self.subscriptions.get(
            publisher, {}).get(table, [])

    # Note(thread-safety): blocking function
    def subscribe_table(self, subscriber, publisher, table):
        """Prepare local service to receives publications from target/table."""
        # data structure: {service -> {target -> set-of-tables}
        LOG.trace("subscribing %s to %s:%s", subscriber, publisher, table)
        if publisher not in self.subscriptions:
            self.subscriptions[publisher] = {}
        if table not in self.subscriptions[publisher]:
            self.subscriptions[publisher][table] = set()
        self.subscriptions[publisher][table].add(subscriber)

        # oslo returns [] instead of set(), so handle that case directly

        # Note(thread-safety): blocking call
        snapshot_seqnum = self.invoke_service_rpc(
            publisher, "get_last_published_data_with_seqnum",
            {'table': table})
        return snapshot_seqnum

    def get_subscription(self, service_id):
        """Return publisher/tables subscribed by service: service_id

        Return data structure:
        {publisher_id -> set of tables}
        """
        result = {}
        for publisher in self.subscriptions:
            for table in self.subscriptions[publisher]:
                if service_id in self.subscriptions[publisher][table]:
                    try:
                        result[publisher].add(table)
                    except KeyError:
                        result[publisher] = set([table])
        return result

    def get_subscribers(self, service_id):
        """List of services subscribed to this service."""

        result = set()
        tables = self.subscriptions.get(service_id, None)
        if not tables:
            # no subscribers
            return []

        for t in tables:
            result = result | self.subscriptions[service_id][t]

        return list(result)

    def to_set_of_tuples(self, snapshot):
        try:
            return set([tuple(x) for x in snapshot])
        except TypeError:
            return snapshot

    def unsubscribe_table(self, subscriber, publisher, table):
        """Remove subscription for local service to target/table."""
        if publisher not in self.subscriptions:
            return False
        if table not in self.subscriptions[publisher]:
            return False
        self.subscriptions[publisher][table].discard(subscriber)
        if len(self.subscriptions[publisher][table]) == 0:
            del self.subscriptions[publisher][table]
        if len(self.subscriptions[publisher]) == 0:
            del self.subscriptions[publisher]

    def _update_tables_with_subscriber(self):
        # not thread-safe: assumes each dseNode is single-threaded
        peers = self.dse_status()['peers']
        for s in self.get_services():
            sid = s.service_id
            # first, include subscriptions within the node, if any
            tables_with_subs = set(self.subscriptions.get(sid, {}))
            # then add subscriptions from other nodes
            for peer_id in peers:
                if sid in peers[peer_id]['subscribed_tables']:
                    tables_with_subs |= peers[
                        peer_id]['subscribed_tables'][sid]
            # call DataService hooks
            if hasattr(s, 'on_first_subs'):
                added = tables_with_subs - s._published_tables_with_subscriber
                if len(added) > 0:
                    s.on_first_subs(added)
            if hasattr(s, 'on_no_subs'):
                removed = \
                    s._published_tables_with_subscriber - tables_with_subs
                if len(removed) > 0:
                    s.on_no_subs(removed)
            s._published_tables_with_subscriber = tables_with_subs

    # Driver CRUD.  Maybe belongs in a subclass of DseNode?
    @classmethod
    def load_drivers(cls):
        """Loads all configured drivers"""
        result = {}
        mgr = stevedore.extension.ExtensionManager(
            namespace='congress.datasource.drivers',
            invoke_on_load=False)

        # Load third party drivers from config if any
        if cfg.CONF.custom_driver_endpoints:
            custom_extensions = cls.load_custom_drivers()
            if custom_extensions:
                mgr.extensions.extend(custom_extensions)

        for driver in mgr:
            if driver.name not in cfg.CONF.disabled_drivers:
                result[driver.name] = driver

        cls.loaded_drivers = result

    @classmethod
    def load_custom_drivers(cls):
        cdist = pkg_resources.get_distribution('congress')
        ext_list = []
        for driver in cfg.CONF.custom_driver_endpoints:
            try:
                ep = pkg_resources.EntryPoint.parse(driver, dist=cdist)
                ep_plugin = ep.load()
                ext = stevedore.extension.Extension(
                    name=ep.name, entry_point=ep, plugin=ep_plugin, obj=None)
                ext_list.append(ext)
            except Exception:
                LOG.exception("Failed to load driver endpoint %s", driver)
        return ext_list

    @classmethod
    def get_driver_info(cls, driver_name):
        driver = cls.loaded_drivers.get(driver_name)
        if not driver:
            raise exception.DriverNotFound(id=driver_name)
        return driver.plugin.get_datasource_info()

    @classmethod
    def get_drivers_info(cls):
        drivers = cls.loaded_drivers.values()
        return [d.plugin.get_datasource_info() for d in drivers]

    @classmethod
    def get_driver_schema(cls, drivername):
        driver = cls.loaded_drivers.get(drivername)
        return driver.plugin.get_schema()

    # Datasource CRUD.  Maybe belongs in a subclass of DseNode?
    # Note(thread-safety): blocking function
    def get_datasource(cls, id_):
        """Return the created datasource."""
        # Note(thread-safety): blocking call
        result = datasources_db.get_datasource(id_)
        if not result:
            raise exception.DatasourceNotFound(id=id_)
        return cls.make_datasource_dict(result)

    # Note(thread-safety): blocking function
    def get_datasources(self, filter_secret=False):
        """Return the created datasources as recorded in the DB.

        This returns what datasources the database contains, not the
        datasources that this server instance is running.
        """
        results = []
        for datasource in datasources_db.get_datasources():
            result = self.make_datasource_dict(datasource)
            if filter_secret:
                # driver_info knows which fields should be secret
                driver_info = self.get_driver_info(result['driver'])
                try:
                    for hide_field in driver_info['secret']:
                        result['config'][hide_field] = "<hidden>"
                except KeyError:
                    pass
            results.append(result)
        return results

    def delete_missing_driver_datasources(self):
        removed = 0
        for datasource in datasources_db.get_datasources():
            try:
                self.get_driver_info(datasource.driver)
            except exception.DriverNotFound:
                datasources_db.delete_datasource_with_data(datasource.id)
                removed = removed+1
                LOG.debug("Datasource driver '%s' not found, deleting the "
                          "datasource '%s' from DB ", datasource.driver,
                          datasource.name)

        LOG.info("Datsource cleanup completed, removed %d datasources",
                 removed)

    def make_datasource_dict(self, req, fields=None):
        result = {'id': req.get('id') or uuidutils.generate_uuid(),
                  'name': req.get('name'),
                  'driver': req.get('driver'),
                  'description': req.get('description'),
                  'type': None,
                  'enabled': req.get('enabled', True)}
        # NOTE(arosen): we store the config as a string in the db so
        # here we serialize it back when returning it.
        if isinstance(req.get('config'), six.string_types):
            result['config'] = json.loads(req['config'])
        else:
            result['config'] = req.get('config')

        return self._fields(result, fields)

    def _fields(self, resource, fields):
        if fields:
            return dict(((key, item) for key, item in resource.items()
                         if key in fields))
        return resource

    def validate_create_datasource(self, req):
        name = req['name']
        if not datalog_compile.string_is_servicename(name):
            raise exception.InvalidDatasourceName(value=name)
        driver = req['driver']
        config = req['config'] or {}

        try:
            loaded_driver = self.get_driver_info(driver)
        except exception.DriverNotFound:
            raise exception.InvalidDriver(driver=req)

        specified_options = set(config.keys())
        valid_options = set(loaded_driver['config'].keys())
        # Check that all the specified options passed in are
        # valid configuration options that the driver exposes.
        invalid_options = specified_options - valid_options
        if invalid_options:
            raise exception.InvalidDriverOption(
                invalid_options=invalid_options)

        # check that all the required options are passed in
        required_options = set(
            [k for k, v in loaded_driver['config'].items()
             if v == constants.REQUIRED])
        missing_options = required_options - specified_options

        if ('project_name' in missing_options and 'tenant_name' in
                specified_options):
            LOG.warning("tenant_name is deprecated, use project_name instead")
            missing_options.remove('project_name')

        if missing_options:
            missing_options = ', '.join(missing_options)
            raise exception.MissingRequiredConfigOptions(
                missing_options=missing_options)
        return loaded_driver

    # Note (thread-safety): blocking function
    def create_datasource_service(self, datasource):
        """Create a new DataService on this node.

        :param: datasource: datsource object.
        """
        # get the driver info for the datasource
        ds_dict = self.make_datasource_dict(datasource)
        if not ds_dict['enabled']:
            LOG.info("datasource %s not enabled, skip loading",
                     ds_dict['name'])
            return
        driver = self.loaded_drivers.get(ds_dict['driver'])
        if not driver:
            raise exception.DriverNotFound(id=ds_dict['driver'])

        if ds_dict['config'] is None:
            args = {'ds_id': ds_dict['id']}
        else:
            args = dict(ds_dict['config'], ds_id=ds_dict['id'])
        kwargs = {'name': ds_dict['name'], 'args': args}
        LOG.info("creating service %s with class %s and args %s",
                 ds_dict['name'], driver.plugin,
                 strutils.mask_password(kwargs, "****"))
        try:
            service = driver.plugin(**kwargs)
        except Exception:
            msg = ("Error loading instance of module '%s'")
            LOG.exception(msg, driver.plugin)
            raise exception.DataServiceError(msg % driver.plugin)
        return service


class DseNodeEndpoints (object):
    """Collection of RPC endpoints that the DseNode exposes on the bus.

       Must be a separate class since all public methods of a given
       class are assumed to be valid RPC endpoints.
    """

    def __init__(self, dsenode):
        self.node = dsenode

    # Note(ekcs): non-sequenced publish retained to simplify rollout of dse2
    #   to be replaced by handle_publish_sequenced
    def handle_publish(self, context, publisher, table, data):
        """Function called on the node when a publication is sent.

           Forwards the publication to all of the relevant services.
        """
        for s in self.node.table_subscribers(publisher, table):
            self.node.service_object(s).receive_data(
                publisher=publisher, table=table, data=data, is_snapshot=True)

    def handle_publish_sequenced(
            self, context, publisher, table, data, is_snapshot, seqnum):
        """Function called on the node when a publication is sent.

           Forwards the publication to all of the relevant services.
        """
        for s in self.node.table_subscribers(publisher, table):
            self.node.service_object(s).receive_data_sequenced(
                publisher=publisher, table=table, data=data, seqnum=seqnum,
                is_snapshot=is_snapshot)



# Copyright (c) 2018 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Mistral Driver for Congress

This driver allows the creation of Congress datasources that interfaces with
Mistral workflows service. The Congress datasource reflects as Congress tables
the Mistral data on workflows, workflow executions, actions, and action
executions. The datasource also supports the triggering of Mistral APIs such as
initiation of a workflows or actions. The triggering of workflows or actions is
especially useful for creating Congress policies that take remedial action.

Datasource creation CLI example:
$ openstack congress datasource create mistral mistral_datasource \
  --config username=admin \
  --config tenant_name=admin \
  --config auth_url=http://127.0.0.1/identity \
  --config password=password
"""

from mistralclient.api.v2 import client as mistral_client

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils


class MistralDriver(datasource_driver.PollingDataSourceDriver,
                    datasource_driver.ExecutionDriver):
    WORKFLOWS = 'workflows'
    ACTIONS = 'actions'

    WORKFLOW_EXECUTIONS = 'workflow_executions'
    ACTION_EXECUTIONS = 'action_executions'

    value_trans = {'translation-type': 'VALUE'}

    workflows_translator = {
        'translation-type': 'HDICT',
        'table-name': WORKFLOWS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'scope', 'translator': value_trans},
             {'fieldname': 'input', 'translator': value_trans},
             {'fieldname': 'namespace', 'translator': value_trans},
             {'fieldname': 'project_id', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans},
             {'fieldname': 'updated_at', 'translator': value_trans},
             {'fieldname': 'definition', 'translator': value_trans},
             {'fieldname': 'description', 'translator': value_trans},
             # TODO(ekcs): maybe enable tags in the future
             )}

    actions_translator = {
        'translation-type': 'HDICT',
        'table-name': ACTIONS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'input', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans},
             {'fieldname': 'updated_at', 'translator': value_trans},
             {'fieldname': 'is_system', 'translator': value_trans},
             {'fieldname': 'definition', 'translator': value_trans},
             {'fieldname': 'description', 'translator': value_trans},
             {'fieldname': 'scope', 'translator': value_trans},
             # TODO(ekcs): maybe enable tags in the future
             )}

    workflow_executions_translator = {
        'translation-type': 'HDICT',
        'table-name': WORKFLOW_EXECUTIONS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'workflow_name', 'translator': value_trans},
             {'fieldname': 'input', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans},
             {'fieldname': 'updated_at', 'translator': value_trans},
             {'fieldname': 'state', 'translator': value_trans},
             {'fieldname': 'state_info', 'translator': value_trans},
             {'fieldname': 'description', 'translator': value_trans},
             {'fieldname': 'workflow_id', 'translator': value_trans},
             {'fieldname': 'workflow_namespace', 'translator': value_trans},
             {'fieldname': 'params', 'translator': value_trans},
             # TODO(ekcs): maybe add task_execution_ids table
             )}

    action_executions_translator = {
        'translation-type': 'HDICT',
        'table-name': ACTION_EXECUTIONS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'state_info', 'translator': value_trans},
             {'fieldname': 'workflow_name', 'translator': value_trans},
             {'fieldname': 'task_execution_id', 'translator': value_trans},
             {'fieldname': 'task_name', 'translator': value_trans},
             {'fieldname': 'description', 'translator': value_trans},
             {'fieldname': 'input', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans},
             {'fieldname': 'updated_at', 'translator': value_trans},
             {'fieldname': 'accepted', 'translator': value_trans},
             {'fieldname': 'state', 'translator': value_trans},
             {'fieldname': 'workflow_namespace', 'translator': value_trans},
             # TODO(ekcs): maybe add action execution tags
             )}

    TRANSLATORS = [
        workflows_translator, actions_translator,
        workflow_executions_translator, action_executions_translator]

    def __init__(self, name='', args=None):
        super(MistralDriver, self).__init__(name, args=args)
        datasource_driver.ExecutionDriver.__init__(self)
        session = ds_utils.get_keystone_session(args)
        self.mistral_client = mistral_client.Client(session=session)

        self.add_executable_client_methods(
            self.mistral_client, 'mistralclient.api.v2.')
        self.initialize_update_method()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'mistral'
        result['description'] = ('Datasource driver that interfaces with '
                                 'Mistral.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['lazy_tables'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def initialize_update_method(self):
        workflows_method = lambda: self._translate_workflows(
            self.mistral_client.workflows.list())
        self.add_update_method(workflows_method, self.workflows_translator)

        workflow_executions_method = (
            lambda: self._translate_workflow_executions(
                self.mistral_client.executions.list()))
        self.add_update_method(workflow_executions_method,
                               self.workflow_executions_translator)

        actions_method = lambda: self._translate_actions(
            self.mistral_client.actions.list())
        self.add_update_method(actions_method, self.actions_translator)

        action_executions_method = lambda: self._translate_action_executions(
            self.mistral_client.action_executions.list())
        self.add_update_method(action_executions_method,
                               self.action_executions_translator)

    @ds_utils.update_state_on_changed(WORKFLOWS)
    def _translate_workflows(self, obj):
        """Translate the workflows represented by OBJ into tables."""
        row_data = MistralDriver.convert_objs(obj, self.workflows_translator)
        return row_data

    @ds_utils.update_state_on_changed(ACTIONS)
    def _translate_actions(self, obj):
        """Translate the workflows represented by OBJ into tables."""
        row_data = MistralDriver.convert_objs(obj, self.actions_translator)
        return row_data

    @ds_utils.update_state_on_changed(WORKFLOW_EXECUTIONS)
    def _translate_workflow_executions(self, obj):
        """Translate the workflow_executions represented by OBJ into tables."""
        row_data = MistralDriver.convert_objs(
            obj, self.workflow_executions_translator)
        return row_data

    @ds_utils.update_state_on_changed(ACTION_EXECUTIONS)
    def _translate_action_executions(self, obj):
        """Translate the action_executions represented by OBJ into tables."""
        row_data = MistralDriver.convert_objs(
            obj, self.action_executions_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.mistral_client, action, action_args)



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Support functions for cfg_validator"""
import uuid

from oslo_log import log as logging

from congress.api import base
from congress import exception
from congress import utils

LOG = logging.getLogger(__name__)

#: Topic for RPC between cfg validator driver (client) and the agents (server)
AGENT_TOPIC = 'congress-validator-agent'
NAMESPACE_CONGRESS = uuid.uuid3(
    uuid.NAMESPACE_URL,
    'http://openstack.org/congress/agent')


def compute_hash(*args):
    """computes a hash from the arguments. Not cryptographically strong."""
    inputs = ''.join([str(arg) for arg in args])
    return str(uuid.uuid3(NAMESPACE_CONGRESS, inputs))


def cfg_value_to_congress(value):
    """Sanitize values for congress

    values of log formatting options typically contains
    '%s' etc, which should not be put in datalog
    """
    if isinstance(value, str):
        value = value.replace('%', '')
    if value is None:
        return ''
    return utils.value_to_congress(value)


def add_rule(bus, policy_name, rules):
    "Adds a policy and rules to the engine"
    try:
        policy_metadata = bus.rpc(
            base.ENGINE_SERVICE_ID,
            'persistent_create_policy_with_rules',
            {'policy_rules_obj': {
                "name": policy_name,
                "kind": "nonrecursive",
                "rules": rules}})
        return policy_metadata
    except exception.CongressException as err:
        LOG.error(err)
        return None



# Copyright (c) 2014 VMware
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys

import testtools

import congress.utils as utils


class UtilsTest(testtools.TestCase):

    def test_value_to_congress(self):
        self.assertEqual("abc", utils.value_to_congress("abc"))
        self.assertEqual("True", utils.value_to_congress(True))
        self.assertEqual("False", utils.value_to_congress(False))
        self.assertEqual(0, utils.value_to_congress(0))
        self.assertEqual(1, utils.value_to_congress(1))
        self.assertEqual(123, utils.value_to_congress(123))
        if sys.version < '3':
            self.assertEqual(456.0, utils.value_to_congress(456.0))

    def test_pretty_rule(self):
        test_rule = "\t \n  head(1, 2)\t \n  "
        expected = "head(1, 2)"
        self.assertEqual(utils.pretty_rule(test_rule), expected)

        test_rule = "\t \n  head(1, 2)\t \n  :- \t \n"
        expected = "head(1, 2)"
        self.assertEqual(utils.pretty_rule(test_rule), expected)

        test_rule = ("\t \n server_with_bad_flavor(id)\t \n  :- \t \n  "
                     "nova:servers(id=id,flavor_id=flavor_id), \t \n "
                     "nova:flavors(id=flavor_id, name=flavor), "
                     "not permitted_flavor(flavor)\t \n ")
        expected = ("server_with_bad_flavor(id) :-\n"
                    "  nova:servers(id=id,flavor_id=flavor_id),\n"
                    "  nova:flavors(id=flavor_id, name=flavor),\n"
                    "  not permitted_flavor(flavor)")
        self.assertEqual(utils.pretty_rule(test_rule), expected)



# Copyright (c) 2018 VMware Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.tests.api import base as api_base
from congress.tests import base


class TestWebhookModel(base.SqlTestCase):
    def setUp(self):
        super(TestWebhookModel, self).setUp()
        services = api_base.setup_config()
        self.webhook_model = services['api']['api-webhook']
        self.node = services['node']
        self.data = services['data']

    def test_add_item(self):
        context = {'ds_id': self.data.service_id}
        payload = {'test_payload': 'test_payload'}
        self.webhook_model.add_item(payload, {}, context=context)
        self.assertEqual(self.data.webhook_payload, payload)



# Copyright (c) 2016 NTT All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock

from congress.datasources import doctor_driver
from congress.tests import base


class TestDoctorDriver(base.TestCase):

    def setUp(self):
        super(TestDoctorDriver, self).setUp()
        self.doctor = doctor_driver.DoctorDriver('test-doctor')

    def numbered_string(self, string, number):
        return string + str(number)

    def generate_events_objects(self, row_number):
        objects = []
        for i in range(0, row_number):
            obj = {
                "time": self.numbered_string('time', i),
                "type": self.numbered_string('type', i),
                "details": {
                    "hostname": self.numbered_string('hostname', i),
                    "status": self.numbered_string('status', i),
                    "monitor": self.numbered_string('monitor', i),
                    "monitor_event_id": self.numbered_string('event_id', i),
                    }
                }
            objects.append(obj)
        return objects

    @mock.patch.object(doctor_driver.DoctorDriver, 'publish')
    def test_events_table(self, mocked_publish):
        objs = self.generate_events_objects(3)
        self.doctor.replace_entire_table_data('events', objs)

        self.assertEqual(3, len(self.doctor.state['events']))

        # change elements in state['events'] set to list and sort by id
        sorted_state = sorted(list(self.doctor.state['events']),
                              key=lambda x: x[0])
        for i, row in enumerate(sorted_state):
            self.assertEqual(self.numbered_string('time', i), row[0])
            self.assertEqual(self.numbered_string('type', i), row[1])
            self.assertEqual(self.numbered_string('hostname', i), row[2])
            self.assertEqual(self.numbered_string('status', i), row[3])
            self.assertEqual(self.numbered_string('monitor', i), row[4])
            self.assertEqual(self.numbered_string('event_id', i), row[5])



# Copyright (c) 2018 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.api import api_utils
from congress.api import base
from congress.api import webservice
from congress import exception


class WebhookModel(base.APIModel):
    """Model for handling webhook notifications."""

    def add_item(self, item, params, id_=None, context=None):
        """POST webhook notification.

        :param item: The webhook payload
        :param params: A dict-like object containing parameters
            from the request query string and body.
        :param id_: not used in this case; should be None
        :param context: Key-values providing frame of reference of request
        """
        caller, source_id = api_utils.get_id_from_context(context)
        try:
            args = {'payload': item}
            # Note(thread-safety): blocking call
            self.invoke_rpc(caller, 'process_webhook_notification', args)
        except exception.CongressException as e:
            raise webservice.DataModelException.create(e)



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""policy library

Revision ID: c0125080d572
Revises: aabe895bbd4d
Create Date: 2017-06-21 13:20:14.529313

"""

# revision identifiers, used by Alembic.
revision = 'c0125080d572'
down_revision = 'aabe895bbd4d'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'library_policies',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False, unique=True),
        sa.Column('abbreviation', sa.String(length=5), nullable=False),
        sa.Column('description', sa.Text(), nullable=False),
        sa.Column('kind', sa.Text(), nullable=False),
        sa.Column('rules', sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )


def downgrade():
    op.drop_table('library_policies')



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.api import api_utils
from congress.api import base
from congress.api import webservice
from congress import exception


LOG = logging.getLogger(__name__)


class RowModel(base.APIModel):
    """Model for handling API requests about Rows."""

    # TODO(thinrichs): No rows have IDs right now.  Maybe eventually
    #   could make ID the hash of the row, but then might as well
    #   just make the ID a string repr of the row.  No use case
    #   for it as of now since all rows are read-only.
    # def get_item(self, id_, context=None):
    #     """Retrieve item with id id\_ from model.

    #     Args:
    #         id_: The ID of the item to retrieve
    #         context: Key-values providing frame of reference of request

    #     Returns:
    #          The matching item or None if item with id\_ does not exist.
    #     """

    # Note(thread-safety): blocking function
    def get_items(self, params, context=None):
        """Get items in model.

        :param: params: A dict-like object containing parameters
                    from the request query string and body.
        :param: context: Key-values providing frame of reference of request

        :returns: A dict containing at least a 'results' key whose value is
                 a list of items in the model.  Additional keys set in the
                 dict will also be rendered for the user.
        """
        LOG.info("get_items(context=%s)", context)
        gen_trace = False
        if 'trace' in params and params['trace'].lower() == 'true':
            gen_trace = True

        # Get the caller, it should be either policy or datasource
        # Note(thread-safety): blocking call
        caller, source_id = api_utils.get_id_from_context(context)
        # FIXME(threod-safety): in DSE2, the returned caller can be a
        #   datasource name. But the datasource name may now refer to a new,
        #   unrelated datasource. Causing the rest of this code to operate on
        #   an unintended datasource.
        #   It would have saved us if table_id was an UUID rather than a name,
        #   but it appears that table_id is just another word for tablename.
        #   Fix: check UUID of datasource before operating. Abort if mismatch
        table_id = context['table_id']
        try:
            args = {'table_id': table_id, 'source_id': source_id,
                    'trace': gen_trace}
            if caller is base.ENGINE_SERVICE_ID:
                # allow extra time for row policy engine query
                # Note(thread-safety): blocking call
                result = self.invoke_rpc(
                    caller, 'get_row_data', args,
                    timeout=self.dse_long_timeout)
            else:
                # Note(thread-safety): blocking call
                result = self.invoke_rpc(caller, 'get_row_data', args)
        except exception.CongressException as e:
            m = ("Error occurred while processing source_id '%s' for row "
                 "data of the table '%s'" % (source_id, table_id))
            LOG.debug(m)
            raise webservice.DataModelException.create(e)

        if gen_trace and caller is base.ENGINE_SERVICE_ID:
            # DSE2 returns lists instead of tuples, so correct that.
            results = [{'data': tuple(x['data'])} for x in result[0]]
            return {'results': results,
                    'trace': result[1] or "Not available"}
        else:
            result = [{'data': tuple(x['data'])} for x in result]
            return {'results': result}

    # Note(thread-safety): blocking function
    def replace_items(self, items, params, context=None):
        """Replaces all data in a table.

        :param: id\_: A table id for replacing all row
        :param: items: A data for new rows
        :param: params: A dict-like object containing parameters from
                    request query
        :param: context: Key-values providing frame of reference of request
        :returns: None
        :raises KeyError: table id doesn't exist
        :raises DataModelException: any error occurs during replacing rows.
        """
        LOG.info("replace_items(context=%s)", context)
        # Note(thread-safety): blocking call
        caller, source_id = api_utils.get_id_from_context(context)
        # FIXME(thread-safety): in DSE2, the returned caller can be a
        #   datasource name. But the datasource name may now refer to a new,
        #   unrelated datasource. Causing the rest of this code to operate on
        #   an unintended datasource.
        #   It would have saved us if table_id was an UUID rather than a name,
        #   but it appears that table_id is just another word for tablename.
        #   Fix: check UUID of datasource before operating. Abort if mismatch
        table_id = context['table_id']
        try:
            args = {'table_id': table_id, 'source_id': source_id,
                    'objs': items}
            # Note(thread-safety): blocking call
            self.invoke_rpc(caller, 'replace_entire_table_data', args)
        except exception.CongressException as e:
            LOG.debug("Error occurred while processing updating rows "
                      "for source_id '%s' and table_id '%s'",
                      source_id, table_id, exc_info=True)
            raise webservice.DataModelException.create(e)
        LOG.info("finish replace_items(context=%s)", context)
        LOG.debug("replaced table %s with row items: %s",
                  table_id, str(items))

    # TODO(thinrichs): It makes sense to sometimes allow users to create
    #  a new row for internal data sources.  But since we don't have
    #  those yet all tuples are read-only from the API.

    # def add_item(self, item, id_=None, context=None):
    #     """Add item to model.

    #     Args:
    #         item: The item to add to the model
    #         id_: The ID of the item, or None if an ID should be generated
    #         context: Key-values providing frame of reference of request

    #     Returns:
    #          Tuple of (ID, newly_created_item)

    #     Raises:
    #         KeyError: ID already exists.
    #     """

    # TODO(thinrichs): once we have internal data sources,
    #   add the ability to update a row.  (Or maybe not and implement
    #   via add+delete.)
    # def update_item(self, id_, item, context=None):
    #     """Update item with id\_ with new data.

    #     Args:
    #         id_: The ID of the item to be updated
    #         item: The new item
    #         context: Key-values providing frame of reference of request

    #     Returns:
    #          The updated item.

    #     Raises:
    #         KeyError: Item with specified id\_ not present.
    #     """
    #     # currently a noop since the owner_id cannot be changed
    #     if id_ not in self.items:
    #         raise KeyError("Cannot update item with ID '%s': "
    #                        "ID does not exist")
    #     return item

    # TODO(thinrichs): once we can create, we should be able to delete
    # def delete_item(self, id_, context=None):
        # """Remove item from model.

        # Args:
        #     id_: The ID of the item to be removed
        #     context: Key-values providing frame of reference of request

        # Returns:
        #      The removed item.

        # Raises:
        #     KeyError: Item with specified id\_ not present.
        # """



# Copyright (c) 2017 VMware Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import copy

from congress.api import webservice
from congress.db import db_library_policies
from congress.tests.api import base as api_base
from congress.tests import base


class TestLibraryPolicyModel(base.SqlTestCase):
    def setUp(self):
        super(TestLibraryPolicyModel, self).setUp()

        services = api_base.setup_config()
        self.library_policy_model = services['api']['api-library-policy']
        self.node = services['node']
        self.engine = services['engine']

        # clear the library policies loaded on startup
        db_library_policies.delete_policies()

        self._add_test_policy()

    def _add_test_policy(self):
        test_policy = {
            "name": "test_policy",
            "description": "test policy description",
            "kind": "nonrecursive",
            "abbreviation": "abbr",
            "rules": [{"rule": "p(x) :- q(x)", "comment": "test comment",
                       "name": "test name"},
                      {"rule": "p(x) :- q2(x)", "comment": "test comment2",
                       "name": "test name2"}]
        }
        test_policy_id, obj = self.library_policy_model.add_item(
            test_policy, {})
        test_policy["id"] = test_policy_id

        test_policy2 = {
            "name": "test_policy2",
            "description": "test policy2 description",
            "kind": "nonrecursive",
            "abbreviation": "abbr2",
            "rules": []
        }
        test_policy_id, obj = self.library_policy_model.add_item(
            test_policy2, {})
        test_policy2["id"] = test_policy_id

        self.policy = test_policy
        self.policy2 = test_policy2

        self.policy_metadata = copy.deepcopy(test_policy)
        self.policy2_metadata = copy.deepcopy(test_policy2)
        del self.policy_metadata['rules']
        del self.policy2_metadata['rules']

    def test_get_items(self):
        ret = self.library_policy_model.get_items({})
        self.assertTrue(all(p in ret['results']
                            for p in [self.policy,
                                      self.policy2]))

        ret = self.library_policy_model.get_items({'include_rules': 'False'})
        self.assertTrue(all(p in ret['results']
                            for p in [self.policy_metadata,
                                      self.policy2_metadata]))

    def test_get_items_by_name(self):
        ret = self.library_policy_model.get_items(
            {'name': 'no-such-policy'})
        self.assertEqual(ret['results'], [])

        ret = self.library_policy_model.get_items(
            {'name': self.policy['name']})
        self.assertEqual(ret['results'], [self.policy])

    def test_get_item(self):
        expected_ret = self.policy
        ret = self.library_policy_model.get_item(self.policy["id"], {})
        self.assertEqual(expected_ret, ret)

        ret = self.library_policy_model.get_item(self.policy["id"],
                                                 {'include_rules': 'False'})
        del expected_ret['rules']
        self.assertEqual(expected_ret, ret)

    def test_get_invalid_item(self):
        self.assertRaises(KeyError,
                          self.library_policy_model.get_item,
                          'invalid-id', {})

    def test_add_item(self):
        test = {
            "name": "test",
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "abbr",
            "rules": []
        }
        expected_ret = copy.deepcopy(test)
        del expected_ret['rules']

        policy_id, policy_obj = self.library_policy_model.add_item(test, {})
        test['id'] = policy_id
        self.assertEqual(test, policy_obj)

    def test_add_item_duplicate_name(self):
        test = {
            "name": "test_policy",
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "abbr",
            "rules": []
        }
        # duplicate name allowed
        self.assertRaises(KeyError,
                          self.library_policy_model.add_item, test, {})
        ret = self.library_policy_model.get_items({})
        self.assertEqual(len(ret['results']), 2)

    def test_add_item_with_id(self):
        test = {
            "name": "test",
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "abbr",
            "rules": []
        }

        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.add_item, test, {}, 'id')

    def test_add_item_without_name(self):
        test = {
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "abbr"
        }

        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.add_item, test, {})

    def test_add_item_with_long_abbreviation(self):
        test = {
            "name": "test",
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "123456",
            "rules": []
        }
        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.add_item, test, {})

    def test_replace_item_without_name(self):
        test = {
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "abbr"
        }

        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.replace_item,
                          self.policy['id'], test, {})

    def test_replace_item_with_long_abbreviation(self):
        test = {
            "name": "test",
            "description": "test description",
            "kind": "nonrecursive",
            "abbreviation": "123456",
            "rules": []
        }
        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.replace_item,
                          self.policy['id'], test, {})

    def test_delete_item(self):
        # delete non-existent policy
        self.assertRaises(KeyError, self.library_policy_model.delete_item,
                          'no_such_policy', {})

        # delete existing policy
        expected_ret = self.policy
        policy_id = self.policy['id']

        ret = self.library_policy_model.delete_item(policy_id, {})
        self.assertEqual(expected_ret, ret)
        self.assertRaises(KeyError,
                          self.library_policy_model.get_item,
                          self.policy['id'], {})

    def test_policy_api_model_error(self):
        """Test the policy api model."""

        # policy without name
        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.add_item,
                          {'rules': []}, {})

        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.replace_item,
                          self.policy['id'], {'rules': []}, {})

        # policy with bad name
        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.add_item,
                          {'name': '7*7', 'rules': []}, {})
        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.replace_item,
                          self.policy['id'], {'name': '7*7', 'rules': []}, {})

        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.add_item,
                          {'name': 'p(x) :- q(x)'}, {})
        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.replace_item,
                          self.policy['id'], {'name': 'p(x) :- q(x)'}, {})

        # policy with invalid 'kind'
        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.add_item,
                          {'kind': 'nonexistent', 'name': 'alice',
                           'rules': []}, {})
        self.assertRaises(webservice.DataModelException,
                          self.library_policy_model.replace_item,
                          self.policy['id'],
                          {'kind': 'nonexistent', 'name': 'alice',
                           'rules': []}, {})

    def test_replace_item(self):
        replacement_policy = {
            "name": "new_name",
            "description": "new test policy2 description",
            "kind": "nonrecursive",
            "abbreviation": "newab",
            "rules": [{"rule": "r(x) :- c(x)", "comment": "test comment",
                       "name": "test name"}]
        }

        # update non-existent item
        self.assertRaises(KeyError,
                          self.library_policy_model.replace_item, 'no_such_id',
                          replacement_policy, {}, {})

        # update existing item
        self.library_policy_model.replace_item(
            self.policy2['id'], replacement_policy, {}, {})

        replacement_policy_w_id = copy.deepcopy(replacement_policy)
        replacement_policy_w_id['id'] = self.policy2['id']

        ret = self.library_policy_model.get_items({})
        self.assertEqual(len(ret['results']), 2)
        self.assertTrue(all(p in ret['results']
                            for p in [self.policy,
                                      replacement_policy_w_id]))



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Handling of RPC

Communication with the datasource driver on the config validator agent
"""

from oslo_config import cfg
import oslo_messaging as messaging
from oslo_service import service

from congress.dse2 import dse_node as dse

DRIVER_TOPIC = (dse.DseNode.SERVICE_TOPIC_PREFIX + 'config' + '-'
                + cfg.CONF.dse.bus_id)


class AgentService(service.Service):
    """Definition of the agent service implemented as an RPC endpoint."""
    def __init__(self, topic, endpoints, conf=None):
        super(AgentService, self).__init__()
        self.conf = conf or cfg.CONF

        self.host = self.conf.agent.host
        self.topic = topic
        self.endpoints = endpoints
        self.transport = messaging.get_transport(self.conf)
        self.target = messaging.Target(exchange=dse.DseNode.EXCHANGE,
                                       topic=self.topic,
                                       version=dse.DseNode.RPC_VERSION,
                                       server=self.host)
        self.server = messaging.get_rpc_server(self.transport,
                                               self.target,
                                               self.endpoints,
                                               executor='eventlet')

    def start(self):
        super(AgentService, self).start()
        self.server.start()

    def stop(self, graceful=False):
        self.server.stop()
        super(AgentService, self).stop(graceful)


class ValidatorDriverClient(object):
    """RPC Proxy used by the agent to access the driver."""
    def __init__(self, topic=DRIVER_TOPIC):
        transport = messaging.get_transport(cfg.CONF)
        target = messaging.Target(exchange=dse.DseNode.EXCHANGE,
                                  topic=topic,
                                  version=dse.DseNode.RPC_VERSION)
        self.client = messaging.RPCClient(transport, target)

    # block calling thread
    def process_templates_hashes(self, context, hashes, host):
        """Sends a list of template hashes to the driver for processing

        :param hashes: the array of hashes
        :param host: the host they come from.
        """
        cctx = self.client.prepare()
        return cctx.call(context, 'process_templates_hashes', hashes=hashes,
                         host=host)

    # block calling thread
    def process_configs_hashes(self, context, hashes, host):
        """Sends a list of config files hashes to the driver for processing

        :param hashes: the array of hashes
        :param host: the host they come from.
        """
        cctx = self.client.prepare()
        return cctx.call(context, 'process_configs_hashes',
                         hashes=hashes, host=host)



# Copyright 2018 - VMware Inc
# Copyright 2016 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import functools

from oslo_db import exception as db_exc
from oslo_log import log as logging
import tenacity

LOG = logging.getLogger(__name__)

_RETRY_ERRORS = (db_exc.DBDeadlock,)


def retry_on_db_error(func):
    """Decorates the given function so that it retries on DB deadlock errors.

    :param func: Function to decorate.
    :return: Decorated function.
    """
    @functools.wraps(func)
    @tenacity.retry(
        reraise=True,
        retry=tenacity.retry_if_exception_type(_RETRY_ERRORS),
        stop=tenacity.stop_after_attempt(10),
        wait=tenacity.wait_incrementing(start=0, increment=0.1, max=2)
    )
    def decorate(*args, **kw):
        try:
            return func(*args, **kw)
        except db_exc.DBDeadlock:
            LOG.exception(
                "DB error detected, operation will be retried: %s", func)
            raise
    return decorate



# Copyright (c) 2018 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock
import sys

sys.modules['mistralclient.api.v2.client'] = mock.Mock()
sys.modules['mistralclient.api.v2'] = mock.Mock()

from congress.datasources import mistral_driver
from congress.tests import base
from congress.tests.datasources import util
from congress.tests import helper

ResponseObj = util.ResponseObj


class TestMistralDriver(base.TestCase):

    def setUp(self):
        super(TestMistralDriver, self).setUp()
        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        args['client'] = mock.MagicMock()
        self.driver = mistral_driver.MistralDriver(
            name='testmistral', args=args)

    def test_list_workflows(self):
        raw_data = [
            ResponseObj({u'created_at': u'2017-10-12 20:06:58',
                         u'definition':
                             u'---\nversion: \'2.0\'\n\nstd.create_instance:\n'
                             u'...',
                         u'id': u'31c429eb-c439-43ec-a633-45c4e8749261',
                         u'input': u'name, image_id, flavor_id, '
                                   u'ssh_username=None, ssh_password=None, '
                                   u'key_name=None, security_groups=None, '
                                   u'nics=None',
                         u'name': u'std.create_instance',
                         u'namespace': u'',
                         u'project_id': u'<default-project>',
                         u'scope': u'public',
                         u'tags': ['tag1', 'tag2'],
                         u'updated_at': None}),
            ResponseObj({u'created_at': u'2017-10-12 20:06:58',
                         u'definition':
                             u'---\nversion: "2.0"\n\nstd.delete_instance:\n'
                             u'...',
                         u'id': u'55f43e39-89aa-43e6-9eec-526b5aa932b9',
                         u'input': u'instance_id',
                         u'name': u'std.delete_instance',
                         u'namespace': u'',
                         u'project_id': u'<default-project>',
                         u'scope': u'public',
                         u'tags': [],
                         u'updated_at': None})]

        translated_data = self.driver._translate_workflows(raw_data)
        self.assertIsNotNone(translated_data)
        self.assertEqual(2, len(translated_data))

        self.assertEqual({
            (u'std.create_instance',
             u'31c429eb-c439-43ec-a633-45c4e8749261',
             u'public',
             u'name, image_id, flavor_id, ssh_username=None, ssh_password='
             u'None, key_name=None, security_groups=None, nics=None',
             u'',
             u'<default-project>',
             u'2017-10-12 20:06:58',
             None,
             u"---\nversion: '2.0'\n\nstd.create_instance:\n...",
             None),
            (u'std.delete_instance',
             u'55f43e39-89aa-43e6-9eec-526b5aa932b9',
             u'public',
             u'instance_id',
             u'',
             u'<default-project>',
             u'2017-10-12 20:06:58',
             None,
             u'---\nversion: "2.0"\n\nstd.delete_instance:\n...',
             None)},
            self.driver.state['workflows'])

    def test_list_actions(self):
        raw_data = [
            ResponseObj({
                u'created_at': u'2017-10-12 20:06:56',
                u'definition': None,
                u'description': u'Updates a load balancer health monitor.',
                u'id': u'f794925d-ed65-41d4-a68d-076412d6ce9d',
                u'input': u'health_monitor, action_region="", body=null',
                u'is_system': True,
                u'name': u'neutron.update_health_monitor',
                u'scope': u'public',
                u'tags': None,
                u'updated_at': None}),
            ResponseObj({
                u'created_at': u'2017-10-13 20:06:56',
                u'definition': u'action definition',
                u'description': u'Updates a load balancer health monitor.',
                u'id': u'a794925d-ed65-41d4-a68d-076412d6ce9d',
                u'input': u'health_monitor, action_region="", body=null',
                u'is_system': False,
                u'name': u'neutron.custom_action',
                u'scope': u'public',
                u'tags': ['tag1', 'tag2'],
                u'updated_at': u'2017-10-13 23:06:56'})]

        translated_data = self.driver._translate_actions(raw_data)
        self.assertIsNotNone(translated_data)
        self.assertEqual(2, len(translated_data))

        self.assertEqual({(u'a794925d-ed65-41d4-a68d-076412d6ce9d',
                           u'neutron.custom_action',
                           u'health_monitor, action_region="", body=null',
                           u'2017-10-13 20:06:56',
                           u'2017-10-13 23:06:56',
                           False,
                           u'action definition',
                           u'Updates a load balancer health monitor.',
                           u'public'),
                          (u'f794925d-ed65-41d4-a68d-076412d6ce9d',
                           u'neutron.update_health_monitor',
                           u'health_monitor, action_region="", body=null',
                           u'2017-10-12 20:06:56',
                           None,
                           True,
                           None,
                           u'Updates a load balancer health monitor.',
                           u'public')},
                         self.driver.state['actions'])

    def test_list_workflow_executions(self):
        raw_data = [
            ResponseObj({u'created_at': u'2017-12-19 22:56:50',
                         u'description': u'',
                         u'id': u'46bbba4b-8a2e-4281-be61-1e92ebfdd6b6',
                         u'input': u'{"instance_id": 1}',
                         u'params': u'{"namespace": "", "task_name": ""}',
                         u'state': u'ERROR',
                         u'state_info': u"Failure caused by error ...",
                         u'task_execution_id': None,
                         u'updated_at': u'2017-12-19 22:57:00',
                         u'workflow_id':
                             u'55f43e39-89aa-43e6-9eec-526b5aa932b9',
                         u'workflow_name': u'std.delete_instance',
                         u'workflow_namespace': u''})]

        translated_data = self.driver._translate_workflow_executions(raw_data)
        self.assertIsNotNone(translated_data)
        self.assertEqual(1, len(translated_data))

        self.assertEqual({(u'46bbba4b-8a2e-4281-be61-1e92ebfdd6b6',
                           u'std.delete_instance',
                           u'{"instance_id": 1}',
                           u'2017-12-19 22:56:50',
                           u'2017-12-19 22:57:00',
                           u'ERROR',
                           u'Failure caused by error ...',
                           u'',
                           u'55f43e39-89aa-43e6-9eec-526b5aa932b9',
                           u'',
                           u'{"namespace": "", "task_name": ""}')},
                         self.driver.state['workflow_executions'])

    def test_list_action_executions(self):
        raw_data = [
            ResponseObj({u'accepted': True,
                         u'created_at': u'2017-12-19 22:56:50',
                         u'description': u'',
                         u'id': u'5c377055-5590-479a-beec-3d4a47a2dfb6',
                         u'input': u'{"server": 1}',
                         u'name': u'nova.servers_delete',
                         u'state': u'ERROR',
                         u'state_info': None,
                         u'tags': None,
                         u'task_execution_id':
                             u'f40a0a20-958d-4948-b0c0-e1961649f2e2',
                         u'task_name': u'delete_vm',
                         u'updated_at': u'2017-12-19 22:56:50',
                         u'workflow_name': u'std.delete_instance',
                         u'workflow_namespace': u''})]

        translated_data = self.driver._translate_action_executions(raw_data)
        self.assertIsNotNone(translated_data)
        self.assertEqual(1, len(translated_data))

        self.assertEqual({(u'5c377055-5590-479a-beec-3d4a47a2dfb6',
                           u'nova.servers_delete',
                           None,
                           u'std.delete_instance',
                           u'f40a0a20-958d-4948-b0c0-e1961649f2e2',
                           u'delete_vm',
                           u'',
                           u'{"server": 1}',
                           u'2017-12-19 22:56:50',
                           u'2017-12-19 22:56:50',
                           True,
                           u'ERROR',
                           u'')},
                         self.driver.state['action_executions'])

    def test_execute(self):
        class MockClient(object):
            def __init__(self):
                self.testkey = None

            def mock_action(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        mock_client = MockClient()
        self.driver.mistral_client = mock_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('mock_action', api_args)

        self.assertEqual(expected_ans, mock_client.testkey)



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg

from congress.api import webservice
from congress.tests.api import base as api_base
from congress.tests import base
from congress.tests import helper


class TestDriverModel(base.SqlTestCase):
    def setUp(self):
        super(TestDriverModel, self).setUp()
        services = api_base.setup_config()
        self.node = services['node']
        self.driver_model = services['api']['api-system']

    def tearDown(self):
        super(TestDriverModel, self).tearDown()

    def test_drivers_list(self):
        context = {}
        drivers = helper.supported_drivers()
        expected_ret = sorted(drivers, key=lambda d: d['id'])
        ret = self.driver_model.get_items({}, context)['results']
        actual_ret = sorted(ret, key=lambda d: d['id'])
        self.assertEqual(expected_ret, actual_ret)

    def test_drivers_list_with_disabled_drivers(self):
        cfg.CONF.set_override('disabled_drivers', 'plexxi')
        services = api_base.setup_config(node_id='test-node-1')
        driver_api = services['api']['api-system']
        drivers = [d['id'] for d in helper.supported_drivers()]
        drivers.remove('plexxi')
        ret = [d['id'] for d in driver_api.get_items({}, {})['results']]
        self.assertEqual(sorted(drivers), sorted(ret))

    def test_drivers_list_with_custom_drivers(self):
        cfg.CONF.set_override(
            'custom_driver_endpoints',
            'test=congress.tests.test_custom_driver:TestCustomDriver')
        services = api_base.setup_config(node_id='test-node-2')
        driver_api = services['api']['api-system']
        ret = [d['id'] for d in driver_api.get_items({}, {})['results']]
        self.assertIn('test', ret)

    def test_driver_details(self):
        context = {
            "driver_id": "fake_datasource"
        }
        expected_ret = {
            "config": {
                "auth_url": "required",
                "endpoint": "(optional)",
                "password": "required",
                "poll_time": "(optional)",
                "region": "(optional)",
                'project_domain_name': '(optional)',
                'user_domain_name': '(optional)',
                "project_name": "required",
                "tenant_name": "(optional)",
                "username": "required"
            },
            "description": "This is a fake driver used for testing",
            "id": "fake_datasource",
            "secret": ["password"],
            "tables": [{'columns': [
                {'description': None, 'name': 'id'},
                {'description': None, 'name': 'name'}],
                'table_id': 'fake_table'}
            ]
        }

        ret = self.driver_model.get_item('fake_datasource', {}, context)
        self.assertEqual(expected_ret, ret)

    def test_invalid_driver_details(self):
        context = {
            "driver_id": "invalid-id"
        }
        self.assertRaises(webservice.DataModelException,
                          self.driver_model.get_item,
                          'invalid-id', {}, context)



# NOTE(ekcs): monkey_patch upfront to ensure all imports get patched modules
import eventlet
# NOTE(ekcs): get_hub() before monkey_patch() to workaround issue with
# import cycles in eventlet < 0.22.0;
# Based on the worked-around in eventlet with patch
# https://github.com/eventlet/eventlet/commit/b756447bab51046dfc6f1e0e299cc997ab343701
# For details please check https://bugs.launchpad.net/congress/+bug/1746136
eventlet.hubs.get_hub()
eventlet.monkey_patch()



#    Copyright 2018 Orange
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for z3types"""
import mock

from congress import data_types
from congress import exception
from congress.tests import base
from congress.tests.z3 import z3mock as z3
from congress.z3 import z3types


def mockz3(f):
    z3types.Z3_AVAILABLE = True
    return mock.patch("congress.z3.z3types.z3", new=z3)(f)


class TestBoolType(base.TestCase):
    """Z3 Boolean values"""

    @mockz3
    def setUp(self):
        self.type = z3types.BoolType()
        super(TestBoolType, self).setUp()

    @mockz3
    def test_to_z3(self):
        self.assertEqual(z3.Z3_OP_TRUE, self.type.to_z3(True).decl().kind())
        self.assertEqual(z3.Z3_OP_FALSE, self.type.to_z3(False).decl().kind())

    @mockz3
    def test_from_z3(self):
        self.assertEqual(True, self.type.to_os(z3.BoolVal(True)))
        self.assertEqual(False, self.type.to_os(z3.BoolVal(False)))


class TestStringType(base.TestCase):
    """Z3 String values"""

    @mockz3
    def setUp(self):
        self.type = z3types.StringType('string', size=8)
        super(TestStringType, self).setUp()

    @mockz3
    def test_to_z3(self):
        x = self.type.to_z3('aaaa')
        self.assertEqual(8, x.size())
        # Do not try to use equality on Z3 values.
        self.assertEqual(self.type.to_z3('aaaa').as_long(), x.as_long())
        self.assertIs(False,
                      self.type.to_z3('bbbb').as_long() == x.as_long())

    @mockz3
    def test_from_z3(self):
        self.assertEqual('aaaa', self.type.to_os(self.type.to_z3('aaaa')))


class TestIntType(base.TestCase):
    """Z3 Int values"""

    @mockz3
    def setUp(self):
        self.type = z3types.IntType('int', size=16)
        super(TestIntType, self).setUp()

    @mockz3
    def test_to_z3(self):
        x = self.type.to_z3(342)
        self.assertEqual(16, x.size())
        self.assertEqual(342, x.as_long())

    @mockz3
    def test_from_z3(self):
        self.assertEqual(421, self.type.to_os(self.type.to_z3(421)))


class TestZ3TypeRegistry(base.TestCase):
    """Other Z3Types Unit tests"""
    def setUp(self):
        try:
            data_types.TypesRegistry.type_class('Foo')
        except KeyError:
            typ = data_types.create_congress_enum_type(
                'Foo', ['a', 'b'], data_types.Str)
            data_types.TypesRegistry.register(typ)
        super(TestZ3TypeRegistry, self).setUp()

    @mockz3
    def test_get_type(self):
        registry = z3types.TypeRegistry()
        t = registry.get_type('Bool')
        self.assertIs(True, isinstance(t, z3.BoolSort))
        t = registry.get_type('Foo')
        self.assertIs(True, isinstance(t, z3.BitVecSort))

    @mockz3
    def test_get_translator(self):
        registry = z3types.TypeRegistry()
        t = registry.get_translator('Bool')
        self.assertIs(True, isinstance(t, z3types.BoolType))
        t = registry.get_translator('Foo')
        self.assertIs(True, isinstance(t, z3types.FiniteType))

    @mockz3
    def test_register(self):
        registry = z3types.TypeRegistry()
        test_type = mock.MagicMock()
        test_type.name = 'Bar'
        registry.register(test_type)
        self.assertEqual(test_type, registry.get_translator('Bar'))

    @mockz3
    def test_reset(self):
        registry = z3types.TypeRegistry()
        test_type = mock.MagicMock()
        test_type.name = 'Bar'
        registry.register(test_type)
        registry.reset()
        test_type.reset.assert_called_once_with()


class TestZ3ToArray(base.TestCase):

    @staticmethod
    def project(z3res):
        return [[x.as_long() for x in vec] for vec in z3res]

    @mockz3
    def test_sat(self):
        self.assertIs(True, z3types.z3_to_array(z3.BoolVal(True)))
        self.assertIs(False, z3types.z3_to_array(z3.BoolVal(False)))

    @mockz3
    def test_simple(self):
        s = z3.BitVecSort(10)

        def vec(x):
            return z3.BitVecVal(x, s)

        x, y = z3.Const('x', s), z3.Const('y', s)
        expr = z3.Eq(x, vec(1))
        self.assertEqual([[1]], self.project(z3types.z3_to_array(expr)))
        expr = z3.And(z3.Eq(x, vec(1)), z3.Eq(y, vec(2)))
        self.assertEqual([[1, 2]], self.project(z3types.z3_to_array(expr)))

    @mockz3
    def test_two_dims(self):
        s = z3.BitVecSort(10)

        def vec(x):
            return z3.BitVecVal(x, s)

        x, y = z3.Const('x', s), z3.Const('y', s)
        expr = z3.Or(z3.Eq(x, vec(1)), z3.Eq(x, vec(2)))
        self.assertEqual([[1], [2]], self.project(z3types.z3_to_array(expr)))

        expr = z3.Or(z3.And(z3.Eq(x, vec(1)), z3.Eq(y, vec(2))),
                     z3.And(z3.Eq(x, vec(3)), z3.Eq(y, vec(4))))
        self.assertEqual([[1, 2], [3, 4]],
                         self.project(z3types.z3_to_array(expr)))

    @mockz3
    def test_fails(self):
        s = z3.BitVecSort(10)
        x = z3.Const('x', s)
        expr = x
        self.assertRaises(exception.PolicyRuntimeException,
                          z3types.z3_to_array, expr)
        expr = z3.Or(x, x)
        self.assertRaises(exception.PolicyRuntimeException,
                          z3types.z3_to_array, expr)



# Copyright (c) 2018 VMware
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import testtools

from congress import data_types


class TestDataTypes(testtools.TestCase):

    def test_nullable(self):
        for type_class in data_types.TYPES:
            self.assertIsNone(type_class.marshal(None))

    def test_Scalar(self):
        valid_values = [1, 1.0, 'str', u'str', True]
        invalid_values = [{}, []]
        for val in valid_values:
            self.assertEqual(val, data_types.Scalar.marshal(val))

        for val in invalid_values:
            self.assertRaises(ValueError, data_types.Scalar.marshal, val)

    def test_Str(self):
        valid_values = ['str', u'str', '1']
        invalid_values = [{}, [], True, 1]
        for val in valid_values:
            self.assertEqual(val, data_types.Str.marshal(val))

        for val in invalid_values:
            self.assertRaises(ValueError, data_types.Str.marshal, val)

    def test_Bool(self):
        valid_values = [True, False]
        invalid_values = [{}, [], 'True', 0, 1]
        for val in valid_values:
            self.assertEqual(val, data_types.Bool.marshal(val))

        for val in invalid_values:
            self.assertRaises(ValueError, data_types.Bool.marshal, val)

    def test_Int(self):
        valid_values = [1, 1.0, -1, True, False]
        invalid_values = [{}, [], 1.1, '1']
        for val in valid_values:
            self.assertEqual(val, data_types.Int.marshal(val))

        for val in invalid_values:
            self.assertRaises(ValueError, data_types.Int.marshal, val)

    def test_Float(self):
        valid_values = [1, 1.0, -1, True, False]
        invalid_values = [{}, [], '1']
        for val in valid_values:
            self.assertEqual(val, data_types.Int.marshal(val))

        for val in invalid_values:
            self.assertRaises(ValueError, data_types.Int.marshal, val)

    def test_UUID(self):
        valid_values = ['026f66d3-d6f1-44d1-8451-0a95ee984ffa',
                        '026f66d3d6f144d184510a95ee984ffa',
                        '-0-2-6f66d3d6f144d184510a95ee984ffa']
        invalid_values = [{}, [], '1', True, 1,
                          'z26f66d3d6f144d184510a95ee984ffa']
        for val in valid_values:
            self.assertEqual(val, data_types.UUID.marshal(val))

        for val in invalid_values:
            self.assertRaises(ValueError, data_types.UUID.marshal, val)

    def test_IPAddress(self):
        valid_values = [('10.0.0.1', '10.0.0.1'),
                        ('::ffff:0a00:0001', '10.0.0.1'),
                        ('0000:0000:0000:0000:0000:ffff:0a00:0001',
                         '10.0.0.1'),
                        ('2001:db8::ff00:42:8329', '2001:db8::ff00:42:8329'),
                        ('2001:0db8:0000:0000:0000:ff00:0042:8329',
                         '2001:db8::ff00:42:8329')]
        invalid_values = [{}, [], '1', True, 1,
                          '256.0.0.1']
        for val in valid_values:
            self.assertEqual(val[1], data_types.IPAddress.marshal(val[0]))

        for val in invalid_values:
            self.assertRaises(ValueError, data_types.IPAddress.marshal, val)

    def test_IPNetwork(self):
        valid_values = [('10.0.0.0/16', '10.0.0.0/16'),
                        ('2001:db00::0/24', '2001:db00::/24'),
                        ('::ffff:0a00:0000/128', '::ffff:a00:0/128')]
        invalid_values = [{}, [], '1', True, 1,
                          '10.0.0.0/4'
                          '10.0.0.1/16']
        for val in valid_values:
            self.assertEqual(val[1], data_types.IPNetwork.marshal(val[0]))

        for val in invalid_values:
            self.assertRaises(ValueError, data_types.IPNetwork.marshal, val)

    def test_enum_types(self):
        Test1 = data_types.create_congress_enum_type(
            'Test', [1, 2], data_types.Int)
        self.assertEqual(1, Test1.marshal(1))
        self.assertIsNone(Test1.marshal(None))
        self.assertRaises(ValueError, Test1.marshal, 0)

        Test2 = data_types.create_congress_enum_type(
            'Test', [1, 2], data_types.Int, catch_all_default_value=0)
        self.assertEqual(1, Test2.marshal(1))
        self.assertEqual(0, Test2.marshal(-1))
        self.assertEqual(0, Test2.marshal('blah'))



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Agent is the main entry point for the configuration validator agent.

The agent is executed on the different nodes of the cloud and sends back
configuration values and metadata to the configuration validator datasource
driver.
"""

import json
import os
import sys

from oslo_config import cfg
from oslo_config import generator
from oslo_log import log as logging
from oslo_service import service
import six

from congress.common import config

from congress.cfg_validator.agent import generator as validator_generator
from congress.cfg_validator.agent import opts as validator_opts
from congress.cfg_validator.agent import rpc
from congress.cfg_validator import parsing
from congress.cfg_validator import utils


LOG = logging.getLogger(__name__)


class Config(object):
    """Encapsulates a configuration file and its meta-data.

    Attributes:
        :ivar path: Path to the configuration on the local file system.
        :ivar template: A Template object to use for parsing the configuration.
        :ivar data: The normalized Namespace loaded by oslo-config, contains
            the parsed values.
        :ivar hash: Hash of the configuration file, salted with the hostname
            and the template hash
        :ivar service_name: The associated service name
    """

    # pylint: disable=protected-access

    def __init__(self, path, template, service_name):
        self.path = path
        self.template = template
        self.data = None
        self.hash = None
        self.service = service_name

    def parse(self, host):
        """Parses the config at the path given. Updates data and hash.

        host: the name of the host where the config is. Used for building a
        unique hash.
        """
        namespaces_data = [ns.data for ns in self.template.namespaces]
        conf = parsing.parse_config_file(namespaces_data, self.path)

        Config.sanitize_config(conf)
        self.data = conf._namespace._normalized

        self.hash = utils.compute_hash(host, self.template.hash,
                                       json.dumps(self.data, sort_keys=True))

    @staticmethod
    def sanitize_config(conf):
        """Sanitizes some cfg.ConfigOpts values, given its options meta-data.

        :param conf: A cfg.ConfigOpts object, pre-loaded with its options
            meta-data and with its configurations values.
        """
        normalized = getattr(conf._namespace,
                             '_normalized', None)
        if not normalized:
            return

        normalized = normalized[0]

        # Obfuscate values of options declared secret
        def _sanitize(opt, group_name='DEFAULT'):
            if not opt.secret:
                return
            if group_name not in normalized:
                return
            if opt.name in normalized[group_name]:
                normalized[group_name][opt.name] = ['*' * 4]

        for option in six.itervalues(conf._opts):
            _sanitize(option['opt'])

        for group_name, group in six.iteritems(conf._groups):
            for option in six.itervalues(group._opts):
                _sanitize(option['opt'], group_name)

    def get_info(self):
        """Information on the configuration file.

        :return: a quadruple made of:
          * the hash of the template,
          * the path to the file,
          * the content
          * the service name.
        """
        return {'template': self.template.hash, 'path': self.path,
                'data': self.data, 'service': self.service}


class Namespace(object):
    """Encapsulates a namespace, as defined by oslo-config-generator.

    It contains the actual meta-data of the options. The data is loaded from
    the service source code, by means of oslo-config-generator.

    Attributes:
        name: The name, as used by oslo-config-generator.
        data: The meta-data of the configuration options.
        hash: Hash of the namespace.
    """

    def __init__(self, name):
        self.name = name
        self.data = None
        self.hash = None

    @staticmethod
    def load(name):
        """Loads a namespace from disk

        :param name: the name of namespace to load.
        :return: a fully configured namespace.
        """

        namespace = Namespace(name)

        saved_conf = cfg.CONF
        cfg.CONF = cfg.ConfigOpts()

        try:
            json_data = validator_generator.generate_ns_data(name)
        finally:
            cfg.CONF = saved_conf

        namespace.hash = utils.compute_hash(json_data)
        namespace.data = json.loads(json_data)

        return namespace

    def get_info(self):
        """Information on the namespace

        :return: a tuple containing
         * data: the content of the namespace
         * name: the name of the namespace
        """
        return {'data': self.data, 'name': self.name}


class Template(object):
    """Describes a template, as defined by oslo-config-generator.

    Attributes:
        :ivar name: The name, as used by oslo-config-generator.
        :ivar path: The path to the template configuration file, as defined by
            oslo-config-generator, on the local file system.
        :ivar output_file: The default output path for this template.
        :ivar namespaces: A set of Namespace objects, which make up this
            template.
    """

    # pylint: disable=protected-access

    def __init__(self, path, output_file):
        self.path = path
        self.output_file = output_file
        self.namespaces = []
        self.hash = None

        name = os.path.basename(output_file)
        self.name = os.path.splitext(name)[0] if name.endswith('.sample') \
            else name

    @staticmethod
    def _parse_template_conf(template_path):
        """Parses a template configuration file"""
        conf = cfg.ConfigOpts()

        conf.register_opts(generator._generator_opts)
        conf(['--config-file', template_path])

        return conf.namespace, conf.output_file

    @staticmethod
    def load(template_path):
        """Loads a template configuration file

        :param template_path: path to the template
        :return: a fully configured Template object.
        """
        namespaces, output_file = Template._parse_template_conf(template_path)

        template = Template(template_path, output_file)

        for namespace in namespaces:
            template.namespaces.append(Namespace.load(namespace))

        template.hash = utils.compute_hash(
            sorted([ns.hash for ns in template.namespaces]))

        return template

    def get_info(self):
        """Info on the template

        :return: a quadruple made of:
            * path: the path to the template path
            * name: the name of the template
            * output_fle:
            * namespaces: an array of namespace hashes.
        """
        return {'path': self.path, 'name': self.name,
                'output_file': self.output_file,
                'namespaces': [ns.hash for ns in self.namespaces]}


class ConfigManager(object):
    """Manages the services configuration files on a node and their meta-data.

    Attributes:
        :ivar host: A hostname.
        :ivar configs: A dict mapping config hashes to their associated Config
            object.
        :ivar templates: A dict mapping template hashes to their associated
            Template object.
        :ivar namespaces: A dict mapping namespace hashes to their associated
            Namespace object.
    """

    def __init__(self, host, services_files):
        self.host = host

        self.configs = {}
        self.templates = {}
        self.namespaces = {}

        for service_name, files in six.iteritems(services_files):
            self.register_service(service_name, files)

    def get_template_by_path(self, template_path):
        """Given a path finds the corresponding template if it is registered

        :param template_path: the path of the searched template
        :return: None or the template
        """
        for template in six.itervalues(self.templates):
            if template.path == template_path:
                return template

    def add_template(self, template_path):
        """Adds a new template (loads it from path).

        :param template_path: a valid path to the template file.
        """
        template = Template.load(template_path)

        self.templates[template.hash] = template
        self.namespaces.update({ns.hash: ns for ns in template.namespaces})

        return template

    def register_config(self, config_path, template_path, service_name):
        """Register a configuration file and its associated template.

        Template and config are actually parsed and loaded.

        :param config_path: a valid path to the config file.
        :param template_path: a valid path to the template file.
        """
        template = (self.get_template_by_path(template_path)
                    or self.add_template(template_path))

        conf = Config(config_path, template, service_name)

        conf.parse(self.host)

        self.configs[conf.hash] = conf

        LOG.info('{hash: %s, path:%s}' % (conf.hash, conf.path))

    def register_service(self, service_name, files):
        """Register all configs for an identified service.

        Inaccessible files are ignored and files registration pursues.

        :param service_name: The name of the service
        :param files: A dict, mapping a configuration path to
            its associated template path
        """
        for config_path, template_path in six.iteritems(files):
            try:
                self.register_config(config_path, template_path, service_name)
            except (IOError, cfg.ConfigFilesNotFoundError):
                LOG.error(('Error while registering config %s with template'
                           ' %s for service %s') %
                          (config_path, template_path, service_name))


class ValidatorAgentEndpoint(object):
    """Validator Agent.

    It is used as an RPC endpoint.

    Attributes:
        config_manager: ConfigManager object.
        driver_api: RPC client to communicate with the driver.
    """

    # pylint: disable=unused-argument,too-many-instance-attributes

    def __init__(self, conf=None):
        self.conf = conf or cfg.CONF
        validator_conf = self.conf.agent

        self.host = validator_conf.host
        self.version = validator_conf.version
        self.max_delay = validator_conf.max_delay
        self.driver_api = rpc.ValidatorDriverClient()

        self.services = list(validator_conf.services.keys())

        service_files = validator_conf.services
        self.config_manager = ConfigManager(self.host, service_files)

    def publish_configs_hashes(self, context, **kwargs):
        """"Sends back all configuration hashes"""
        LOG.info('Sending config hashes')
        conf = set(self.config_manager.configs)
        self.driver_api.process_configs_hashes({}, conf, self.host)

    def publish_templates_hashes(self, context, **kwargs):
        """"Sends back all template hashes"""
        LOG.info('Sending template hashes')
        tpl = set(self.config_manager.templates)
        self.driver_api.process_templates_hashes({}, tpl, self.host)

    def get_namespace(self, context, **kwargs):
        """"Sends back a namespace

        :param context: the RPC context
        :param hash: the hash of the namespace to send
        :return: the namespace or None if not found
        """
        ns_hash = kwargs['ns_hash']
        LOG.info('Sending namespace %s' % ns_hash)

        namespace = self.config_manager.namespaces.get(ns_hash, None)
        if namespace is None:
            return None
        ret = namespace.get_info()
        ret['version'] = self.version
        return ret

    def get_template(self, context, **kwargs):
        """"Sends back a template

        :param context: the RPC context
        :param hash: the hash of the template to send
        :return: the template or None if not found
        """
        template_hash = kwargs['tpl_hash']

        LOG.info('Sending template %s' % template_hash)

        template = self.config_manager.templates.get(template_hash, None)
        if template is None:
            return None
        ret = template.get_info()
        ret['version'] = self.version
        return ret

    def get_config(self, context, **kwargs):
        """"Sends back a config

        :param context: the RPC context
        :param hash: the hash of the config to send
        :return: the config or None if not found
        """
        config_hash = kwargs['cfg_hash']

        LOG.info('Sending config %s' % config_hash)

        conf = self.config_manager.configs.get(config_hash, None)
        if conf is None:
            return None
        ret = conf.get_info()
        ret['version'] = self.version
        return ret


def main():
    """Agent entry point"""
    validator_opts.register_validator_agent_opts(cfg.CONF)
    config.init(sys.argv[1:])
    config.setup_logging()

    if not cfg.CONF.config_file:
        sys.exit("ERROR: Unable to find configuration file via default "
                 "search paths ~/.congress/, ~/, /etc/congress/, /etc/) and "
                 "the '--config-file' option!")

    agent = ValidatorAgentEndpoint()
    server = rpc.AgentService(utils.AGENT_TOPIC, [agent])
    service.launch(agent.conf, server).wait()


if __name__ == '__main__':
    main()



# Copyright (c) 2017 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json

from oslo_db import exception as oslo_db_exc
import sqlalchemy as sa
from sqlalchemy.orm import exc as db_exc

from congress.db import api as db
from congress.db import model_base
from congress.db import utils as db_utils


class LibraryPolicy(model_base.BASE, model_base.HasId):
    __tablename__ = 'library_policies'

    name = sa.Column(sa.String(255), nullable=False, unique=True)
    abbreviation = sa.Column(sa.String(5), nullable=False)
    description = sa.Column(sa.Text(), nullable=False)
    kind = sa.Column(sa.Text(), nullable=False)
    rules = sa.Column(sa.Text(), nullable=False)

    def to_dict(self, include_rules=True, json_rules=False):
        """From a given library policy, return a policy dict.

        :param: include_rules (bool, optional): include policy rules in return
                dictionary. Defaults to False.
        """
        if not include_rules:
            d = {'id': self.id,
                 'name': self.name,
                 'abbreviation': self.abbreviation,
                 'description': self.description,
                 'kind': self.kind}
        else:
            d = {'id': self.id,
                 'name': self.name,
                 'abbreviation': self.abbreviation,
                 'description': self.description,
                 'kind': self.kind,
                 'rules': (self.rules if json_rules
                           else json.loads(self.rules))}
        return d


@db_utils.retry_on_db_error
def add_policy(policy_dict, session=None):
    session = session or db.get_session()
    try:
        with session.begin(subtransactions=True):
            new_row = LibraryPolicy(
                name=policy_dict['name'],
                abbreviation=policy_dict['abbreviation'],
                description=policy_dict['description'],
                kind=policy_dict['kind'],
                rules=json.dumps(policy_dict['rules']))
            session.add(new_row)
        return new_row
    except oslo_db_exc.DBDuplicateEntry:
        raise KeyError(
            "Policy with name %s already exists" % policy_dict['name'])


@db_utils.retry_on_db_error
def replace_policy(id_, policy_dict, session=None):
    session = session or db.get_session()
    try:
        with session.begin(subtransactions=True):
            new_row = LibraryPolicy(
                id=id_,
                name=policy_dict['name'],
                abbreviation=policy_dict['abbreviation'],
                description=policy_dict['description'],
                kind=policy_dict['kind'],
                rules=json.dumps(policy_dict['rules']))
            session.query(LibraryPolicy).filter(
                LibraryPolicy.id == id_).one().update(
                new_row.to_dict(include_rules=True, json_rules=True))
        return new_row
    except db_exc.NoResultFound:
        raise KeyError('No policy found with policy id %s' % id_)


@db_utils.retry_on_db_error
def delete_policy(id_, session=None):
    session = session or db.get_session()
    return session.query(LibraryPolicy).filter(
        LibraryPolicy.id == id_).delete()


@db_utils.retry_on_db_error
def delete_policies(session=None):
    session = session or db.get_session()
    return session.query(LibraryPolicy).delete()


@db_utils.retry_on_db_error
def get_policy(id_, session=None):
    session = session or db.get_session()
    try:
        return session.query(LibraryPolicy).filter(
            LibraryPolicy.id == id_).one()
    except db_exc.NoResultFound:
        raise KeyError('No policy found with policy id %s' % id_)


@db_utils.retry_on_db_error
def get_policy_by_name(name, session=None):
    session = session or db.get_session()
    try:
        return session.query(LibraryPolicy).filter(
            LibraryPolicy.name == name).one()
    except db_exc.NoResultFound:
        raise KeyError('No policy found with policy name %s' % name)


@db_utils.retry_on_db_error
def get_policies(session=None):
    session = session or db.get_session()
    return (session.query(LibraryPolicy).all())



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json

import sqlalchemy as sa
from sqlalchemy.orm import exc as db_exc

from congress.db import api as db
from congress.db import model_base
from congress.db import utils as db_utils


class DSTableData(model_base.BASE):
    __tablename__ = 'dstabledata'

    ds_id = sa.Column(sa.String(36), nullable=False, primary_key=True)
    tablename = sa.Column(sa.String(255), nullable=False, primary_key=True)
    # choose long length compatible with MySQL, SQLite, Postgres
    tabledata = sa.Column(sa.Text(), nullable=False)


@db_utils.retry_on_db_error
def store_ds_table_data(ds_id, tablename, tabledata, session=None):
    session = session or db.get_session()
    tabledata = _json_encode_table_data(tabledata)
    with session.begin(subtransactions=True):
        new_row = session.merge(DSTableData(
            ds_id=ds_id,
            tablename=tablename,
            tabledata=tabledata))
    return new_row


@db_utils.retry_on_db_error
def delete_ds_table_data(ds_id, tablename=None, session=None):
    session = session or db.get_session()
    if tablename is None:
        return session.query(DSTableData).filter(
            DSTableData.ds_id == ds_id).delete()
    else:
        return session.query(DSTableData).filter(
            DSTableData.ds_id == ds_id,
            DSTableData.tablename == tablename).delete()


@db_utils.retry_on_db_error
def get_ds_table_data(ds_id, tablename=None, session=None):
    session = session or db.get_session()
    try:
        if tablename is None:
            rows = session.query(DSTableData).filter(
                DSTableData.ds_id == ds_id)
            return_list = []
            for row in rows:
                return_list.append(
                    {'tablename': row.tablename,
                     'tabledata': _json_decode_table_data(row.tabledata)})
            return return_list
        else:
            return _json_decode_table_data(session.query(DSTableData).filter(
                DSTableData.ds_id == ds_id,
                DSTableData.tablename == tablename).one().tabledata)
    except db_exc.NoResultFound:
        pass


def _json_encode_table_data(tabledata):
    tabledata = list(tabledata)
    for i in range(0, len(tabledata)):
        tabledata[i] = list(tabledata[i])
    return json.dumps(tabledata)


def _json_decode_table_data(json_tabledata):
    tabledata = json.loads(json_tabledata)
    for i in range(0, len(tabledata)):
        tabledata[i] = tuple(tabledata[i])
    return set(tabledata)
